# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
"""Interface for getting Crash and Bucket data from CrashManager API"""
import json
from contextlib import contextmanager
from logging import getLogger
from os import unlink
from pathlib import Path
from shutil import rmtree
from tempfile import mkdtemp, mkstemp

from Collector.Collector import Collector

from .utils import grz_tmp

LOG = getLogger(__name__)


class Bucket:
    """Get Bucket data for a specified CrashManager bucket."""

    def __init__(self, bucket_id):
        """Initialize a Bucket instance.

        Arguments:
            bucket_id (int): ID of the requested bucket on the server side
        """
        assert isinstance(bucket_id, int)
        self._bucket_id = bucket_id
        self._sig_filename = None
        self._coll = Collector()
        self._url = (
            f"{self._coll.serverProtocol}://{self._coll.serverHost}:"
            f"{self._coll.serverPort}/crashmanager/rest/buckets/{bucket_id}/"
        )
        self._data = None

    @property
    def bucket_id(self):
        return self._bucket_id

    def __enter__(self):
        return self

    def __exit__(self, *exc):
        self.cleanup()

    def __getattr__(self, name):
        if self._data is None:
            self._data = self._coll.get(self._url).json()
        if name not in self._data:
            raise AttributeError(
                f"'{type(self).__name__,}' object has no attribute '{name}'"
                f" (has: {list(self._data)})"
            )
        return self._data[name]

    def __setattr__(self, name, value):
        if name.startswith("_"):
            super().__setattr__(name, value)
            return
        raise AttributeError("can't set attribute")

    def cleanup(self):
        """Cleanup any resources held by this instance.

        Arguments:
            None

        Returns:
            None
        """
        if self._sig_filename is not None:
            rmtree(self._sig_filename.parent)

    def iter_crashes(self, quality_filter=None):
        """Fetch all crash IDs for this FuzzManager bucket.
        Only crashes with testcases are returned.

        Arguments:
            quality_filter (int): Filter crashes by quality value (None for all)

        Returns:
            generator: generator of CrashEntry
        """

        def _get_results(endpoint, params=None):
            """
            Function to get paginated results from FuzzManager

            Args:
                endpoint (str): FuzzManager REST API to query (eg. "crashes").
                params (dict): Params to pass through to requests.get

            Returns:
                generator: objects returned by FuzzManager (as dicts)
            """
            LOG.debug("first request to /%s/", endpoint)

            url = (
                f"{self._coll.serverProtocol}://{self._coll.serverHost}:"
                f"{self._coll.serverPort}/crashmanager/rest/{endpoint}/"
            )

            response = self._coll.get(url, params=params).json()

            while True:
                LOG.debug(
                    "got %d/%d %s",
                    len(response["results"]),
                    response["count"],
                    endpoint,
                )
                while response["results"]:
                    yield response["results"].pop()

                if response["next"] is None:
                    break

                LOG.debug("next request to /%s/", endpoint)
                response = self._coll.get(response["next"]).json()

        # Get all crashes for bucket
        query_args = [
            ("op", "AND"),
            ("bucket", self.bucket_id),
        ]
        if quality_filter is not None:
            query_args.append(("testcase__quality", quality_filter))
        query = json.dumps(dict(query_args))

        n_yielded = 0
        for crash in _get_results(
            "crashes", params={"query": query, "include_raw": "0"}
        ):

            if not crash["testcase"]:
                LOG.warning("crash %d has no testcase, skipping", crash["id"])
                continue

            n_yielded += 1
            LOG.debug("yielding crash #%d", n_yielded)
            result = CrashEntry(crash["id"])
            result._data = crash  # pylint: disable=protected-access
            yield result

    def signature_path(self):
        """Download the bucket data from CrashManager.

        Arguments:
            None

        Returns:
            Path: Path on disk where signature exists.
        """
        if self._sig_filename is not None:
            return self._sig_filename

        tmpd = Path(
            mkdtemp(prefix=f"bucket-{self._bucket_id}-", dir=grz_tmp("fuzzmanager"))
        )
        try:
            sig_basename = f"{self._bucket_id}.signature"
            sig_filename = tmpd / sig_basename
            sig_filename.write_text(self.signature)
            sigmeta_filename = sig_filename.with_suffix(".metadata")
            sigmeta_filename.write_text(
                json.dumps(
                    {
                        "size": self.size,
                        "frequent": self.frequent,
                        "shortDescription": self.shortDescription,
                        "testcase__quality": self.best_quality,
                    }
                )
            )
        except:  # noqa pragma: no cover pylint: disable=bare-except
            rmtree(tmpd)
            raise

        self._sig_filename = sig_filename
        return self._sig_filename


class CrashEntry:
    """Get the CrashEntry data for the specified CrashManager crash.

    Attributes:
        crash_id (int): the server ID for the crash
        see crashmanager.serializers.CrashEntrySerializer
    """

    RAW_FIELDS = frozenset({"rawCrashData", "rawStderr", "rawStdout"})

    def __init__(self, crash_id):
        """Initialize CrashEntry.

        Arguments:
            crash_id (int): ID of the requested crash on the server side
        """
        assert isinstance(crash_id, int)
        self._crash_id = crash_id
        self._coll = Collector()
        self._url = (
            f"{self._coll.serverProtocol}://{self._coll.serverHost}:"
            f"{self._coll.serverPort}/crashmanager/rest/crashes/{crash_id}/"
        )
        self._data = None
        self._tc_filename = None

    @property
    def crash_id(self):
        return self._crash_id

    def __enter__(self):
        return self

    def __exit__(self, *exc):
        self.cleanup()

    def __getattr__(self, name):
        if self._data is None or (name in self.RAW_FIELDS and name not in self._data):
            need_raw = "1" if name in self.RAW_FIELDS else "0"
            # TODO: handle 403 and 404?
            self._data = self._coll.get(
                self._url, params={"include_raw": need_raw}
            ).json()
        if name not in self._data:
            raise AttributeError(
                f"'{type(self).__name__}' object has no attribute '{name}' "
                f"(has: {list(self._data)})"
            )
        return self._data[name]

    def __setattr__(self, name, value):
        if name.startswith("_"):
            super().__setattr__(name, value)
            return
        if name != "testcase_quality":
            raise AttributeError("can't set attribute")
        self._coll.patch(self._url, data={name: value})
        if self._data:
            self._data[name] = value

    def cleanup(self):
        """Cleanup any resources held by this instance.

        Arguments:
            None

        Returns:
            None
        """
        if self._tc_filename is not None:
            self._tc_filename.unlink()

    def testcase_path(self):
        """Download the testcase data from CrashManager.

        Arguments:
            None

        Returns:
            Path: Path on disk where testcase exists_
        """
        if self._tc_filename is not None:
            return self._tc_filename

        dlurl = self._url + "download/"
        response = self._coll.get(dlurl)

        if "content-disposition" not in response.headers:
            raise RuntimeError(
                f"Server sent malformed response: {response!r}"
            )  # pragma: no cover

        handle, filename = mkstemp(
            dir=grz_tmp("fuzzmanager"),
            prefix=f"crash-{self.crash_id}-",
            suffix=Path(self.testcase).suffix,
        )
        try:
            with open(handle, "wb") as output:
                output.write(response.content)
        except:  # noqa pragma: no cover pylint: disable=bare-except
            unlink(filename)
            raise
        self._tc_filename = Path(filename)
        return self._tc_filename


@contextmanager
def load_fm_data(crash_id, load_bucket=False):
    """Load CrashEntry including Bucket from FuzzManager.

    Arguments:
        crash_id (int): Crash ID to load.
        load_bucket (bool): Attempt to load bucket.

    Yields:
        2-tuple(CrashEntry, Bucket): Data loaded from FuzzManager.
    """
    with CrashEntry(crash_id) as crash:
        # load signature if needed
        if load_bucket and crash.bucket:
            with Bucket(crash.bucket) as bucket:
                yield crash, bucket
        else:
            yield crash, None



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
"""Unit tests for `grizzly.replay.crash` and `grizzly.replay.bucket`."""
from copy import deepcopy

from pytest import mark

from ..common.fuzzmanager import Bucket, CrashEntry
from ..common.utils import Exit
from .bucket import bucket_main
from .crash import main as crash_main
from .crash import modify_args


def test_crash_main(mocker):
    """test main()"""
    mocker.patch("grizzly.replay.crash.ReplayManager.main", return_value=0)
    crash = mocker.Mock(spec=CrashEntry, crash_id=1, tool="tool-name")
    load_fm_data = mocker.patch("grizzly.replay.crash.load_fm_data")
    load_fm_data.return_value.__enter__ = mocker.Mock(return_value=(crash, None))
    args = mocker.Mock(input=12345, sig=None, tool=None)
    assert crash_main(args) == 0


@mark.parametrize(
    "arg_tool, signature",
    [
        (None, None),
        ("arg_tool", None),
        (None, "bucket_sig"),
    ],
)
def test_modify_args(mocker, arg_tool, signature):
    """test modify_args()"""
    args = mocker.Mock(input="org_input", tool=arg_tool, sig=None)
    crash = mocker.Mock(spec=CrashEntry, tool="crash_tool")
    crash.testcase_path.return_value = "crash_input"
    if signature is not None:
        bucket = mocker.Mock(spec=Bucket)
        bucket.signature_path.return_value = signature
    else:
        bucket = None
    mod = modify_args(args, crash, bucket)
    assert mod.original_crash_id == "org_input"
    assert mod.input == "crash_input"
    assert mod.tool == ("crash_tool" if arg_tool is None else arg_tool)
    assert mod.sig == signature


@mark.parametrize(
    "crashes, main_exit_codes, result, arg_sig, arg_tool",
    [
        # no crashes -> success
        ([], [], Exit.SUCCESS, None, None),
        # 1 crash fails -> no success
        ([(123, "test-tool")], [Exit.ERROR], Exit.ERROR, None, None),
        # second of 3 succeeds -> success
        (
            [(123, "test-tool"), (456, "test-tool2")],
            [Exit.ERROR, Exit.SUCCESS],
            Exit.SUCCESS,
            None,
            None,
        ),
        # --sig is respected
        ([(123, "test-tool")], [Exit.SUCCESS], Exit.SUCCESS, "test_sig2.json", None),
        # --tool is respected
        ([(123, "test-tool")], [Exit.SUCCESS], Exit.SUCCESS, None, "test-tool-arg"),
        # abort in crash main should also abort bucket main
        ([(123, "test-tool")], [Exit.ABORT], Exit.ABORT, None, None),
    ],
)
def test_bucket_main(mocker, crashes, main_exit_codes, result, arg_sig, arg_tool):
    """tests for `grizzly.replay.bucket.bucket_main`"""
    call_args = []

    def copy_args(args):
        call_args.append(deepcopy(args))
        return main_exit_codes[main.call_count - 1]

    bucket = mocker.Mock(spec=Bucket)
    bucket.signature_path.return_value = "test_sig.json"
    bucket.iter_crashes.return_value = (
        mocker.Mock(crash_id=crash, tool=tool) for crash, tool in crashes
    )
    fake_bucket = mocker.patch("grizzly.replay.bucket.Bucket", autospec=True)
    fake_bucket.return_value.__enter__.return_value = bucket

    main = mocker.Mock(side_effect=copy_args)

    args = mocker.Mock(input=789, sig=arg_sig, tool=arg_tool)
    assert bucket_main(args, main) == result
    assert main.call_count == len(main_exit_codes)
    for idx, (crash, _tool) in enumerate(crashes[: main.call_count]):
        assert call_args[idx].input == crash
        if arg_tool is not None:
            assert call_args[idx].tool == arg_tool
        else:
            assert call_args[idx].tool is None
        if arg_sig is not None:
            assert call_args[idx].sig == arg_sig
        else:
            assert call_args[idx].sig == "test_sig.json"



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.

from logging import getLogger

from ..common.fuzzmanager import load_fm_data
from ..main import configure_logging
from .args import ReplayFuzzManagerIDArgs
from .replay import ReplayManager

LOG = getLogger(__name__)


def main(args):
    """CLI for `grizzly.replay.crash`.

    Arguments:
        args (argparse.Namespace): Result from `ReplayArgs.parse_args`.

    Returns:
        int: 0 for success. non-0 indicates a problem.
    """
    configure_logging(args.log_level)
    with load_fm_data(args.input, load_bucket=not args.sig) as (crash, bucket):
        LOG.info("Loaded crash %d from FuzzManager", crash.crash_id)
        # call grizzly.replay
        return ReplayManager.main(modify_args(args, crash, bucket))


def modify_args(args, crash, bucket):
    """

    Arguments:
        args (argparse.Namespace): Result from `ReplayArgs.parse_args`.



    Returns:
        args (argparse.Namespace): Modified arguments.
    """
    args.original_crash_id = args.input
    args.input = crash.testcase_path()
    if args.tool is None:
        LOG.info("Setting default --tool=%s from CrashEntry", crash.tool)
        args.tool = crash.tool

    # load signature if needed
    if bucket is not None:
        args.sig = bucket.signature_path()

    return args


if __name__ == "__main__":
    raise SystemExit(main(ReplayFuzzManagerIDArgs().parse_args()))



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
from platform import system

from pytest import mark, raises

from .args import CommonArgs, GrizzlyArgs


@mark.parametrize(
    "extra_args, results",
    [
        # test default
        ([], {}),
        # test no-harness and relaunch
        (["--no-harness"], {"relaunch": 1}),
    ],
)
def test_common_args_01(mocker, tmp_path, extra_args, results):
    """test CommonArgs.parse_args() - success"""
    mocker.patch("grizzly.args.scan_plugins", return_value=["targ"])
    fake_bin = tmp_path / "fake.bin"
    fake_bin.touch()
    cmd = [str(fake_bin), "--platform", "targ"] + extra_args
    args = vars(CommonArgs().parse_args(argv=cmd))
    assert args
    for arg_name, expected_value in results.items():
        assert args[arg_name] == expected_value, f"'{arg_name}' does not match"


@mark.parametrize(
    "args, msg, idx",
    [
        # test help
        (["-h"], "For addition help check out the wiki", 0),
        # test without args
        ([], "the following arguments are required: binary", -1),
        # test with missing bin
        (["missing-bin"], "error: file not found: 'missing-bin'", -1),
    ],
)
def test_common_args_02(capsys, mocker, args, msg, idx):
    """test CommonArgs.parse_args()"""
    mocker.patch("grizzly.args.scan_plugins", return_value=["x"])
    with raises(SystemExit):
        CommonArgs().parse_args(argv=args)
    assert msg in capsys.readouterr()[idx]


@mark.parametrize(
    "args, msg, targets",
    [
        # test no installed platforms
        ([], "error: No Platforms (Targets) are installed", []),
        # test invalid log limit
        (["--log-limit", "-1"], "error: --log-limit must be >= 0", ["targ1"]),
        # test invalid memory limit
        (["--memory", "-1"], "error: --memory must be >= 0", ["targ1"]),
        # test --no-harness and --time-limit
        (
            ["--no-harness", "--time-limit", "10"],
            "error: --time-limit cannot be used with --no-harness",
            ["targ1"],
        ),
        # test invalid relaunch value
        (["--relaunch", "0"], "error: --relaunch must be >= 1", ["targ1"]),
        # test invalid asset
        (
            ["--platform", "targ1", "--asset", "bad", "a"],
            "error: Asset 'bad' not supported by target",
            ["targ1"],
        ),
        # test invalid asset path
        (
            ["--platform", "targ1", "--asset", "a", "a"],
            "error: Failed to add asset 'a' cannot find 'a'",
            ["targ1"],
        ),
        # test invalid time-limit
        (
            ["--platform", "targ1", "--time-limit", "-1"],
            "error: --time-limit must be >= 1",
            ["targ1"],
        ),
        # test invalid timeout
        (
            ["--platform", "targ1", "--timeout", "-1"],
            "error: --timeout must be >= 0",
            ["targ1"],
        ),
        # test timeout < time-limit
        (
            ["--platform", "targ1", "--timeout", "1", "--time-limit", "10"],
            "error: --timeout must be >= --time-limit",
            ["targ1"],
        ),
        # test invalid tool usage
        (
            ["--platform", "targ1", "--tool", "x"],
            "error: --tool requires --fuzzmanager",
            ["targ1"],
        ),
        # test invalid launch-attempts value
        (
            ["--launch-attempts", "0"],
            "error: --launch-attempts must be >= 1",
            ["targ1"],
        ),
    ],
)
def test_common_args_03(capsys, mocker, tmp_path, args, msg, targets):
    """test CommonArgs.parse_args()"""
    mocker.patch("grizzly.args.scan_plugins", autospec=True, return_value=targets)
    mocker.patch(
        "grizzly.args.scan_target_assets",
        autospec=True,
        return_value={"targ1": ["a", "b"]},
    )
    mocker.patch("grizzly.args.system", autospec=True, return_value="foo")
    fake_bin = tmp_path / "fake.bin"
    fake_bin.touch()
    with raises(SystemExit):
        CommonArgs().parse_args(argv=[str(fake_bin)] + args)
    assert msg in capsys.readouterr()[-1]


def test_common_args_04(tmp_path):
    """test CommonArgs.parse_args() '--logs' must be dir"""
    fake_bin = tmp_path / "fake.bin"
    fake_bin.touch()
    # test with file
    with raises(SystemExit):
        CommonArgs().parse_args([str(fake_bin), "--logs", str(fake_bin)])
    # test with dir
    CommonArgs().parse_args([str(fake_bin), "--logs", str(tmp_path)])


def test_common_args_05(mocker):
    """test CommonArgs.is_headless()"""
    mocker.patch("grizzly.args.system", autospec=True, return_value="Linux")
    getenv = mocker.patch("grizzly.args.getenv", autospec=True)
    # test not headless
    getenv.return_value = ":0"
    assert not CommonArgs.is_headless()
    # test headless
    getenv.return_value = None
    assert CommonArgs.is_headless()


@mark.parametrize(
    "extra_args, results",
    [
        # test default
        ([], {}),
        # test smoke-test
        (["--smoke-test"], {"limit": 10}),
    ],
)
def test_grizzly_args_01(mocker, tmp_path, extra_args, results):
    """test GrizzlyArgs.parse_args() - success"""
    mocker.patch(
        "grizzly.args.scan_plugins",
        autospec=True,
        side_effect=(["targ"], ["adpt"]),
    )
    fake_bin = tmp_path / "fake.bin"
    fake_bin.touch()
    cmd = [str(fake_bin), "adpt", "--platform", "targ"] + extra_args
    args = vars(GrizzlyArgs().parse_args(argv=cmd))
    assert args
    for arg_name, expected_value in results.items():
        assert args[arg_name] == expected_value, f"'{arg_name}' does not match"


@mark.parametrize(
    "args, msg, idx",
    [
        # test help
        (["-h"], "For addition help check out the wiki", 0),
        # test without args
        ([], "the following arguments are required: binary, adapter", -1),
        # test missing binary
        (["missing-bin", "x"], "error: file not found: 'missing-bin'", -1),
    ],
)
def test_grizzly_args_02(capsys, mocker, args, msg, idx):
    """test GrizzlyArgs.parse_args()"""
    mocker.patch("grizzly.args.scan_plugins", autospec=True, return_value=["x"])
    with raises(SystemExit):
        GrizzlyArgs().parse_args(argv=args)
    assert msg in capsys.readouterr()[idx]


def test_grizzly_args_03(capsys, mocker):
    """test GrizzlyArgs.parse_args() - no adapters installed"""
    mocker.patch("grizzly.args.scan_plugins", side_effect=(["t"], []))
    with raises(SystemExit):
        GrizzlyArgs().parse_args(argv=["b", "a"])
    assert "error: No Adapters are installed" in capsys.readouterr()[-1]


@mark.parametrize(
    "args, msg",
    [
        # test invalid collect value
        (["--collect", "0"], "error: --collect must be greater than 0"),
        # test missing input
        (["--input", "missing"], "error: 'missing' does not exist"),
        # test invalid limit value
        (["--limit", "-1"], "error: --limit must be >= 0"),
        # test invalid limit value
        (["--limit-reports", "-1"], "error: --limit-reports must be >= 0"),
        # test runtime limit value
        (["--runtime", "-1"], "error: --runtime must be >= 0"),
    ],
)
def test_grizzly_args_04(capsys, mocker, tmp_path, args, msg):
    """test CommonArgs.parse_args()"""
    mocker.patch(
        "grizzly.args.scan_plugins",
        autospec=True,
        side_effect=(["targ"], ["adpt"]),
    )
    fake_bin = tmp_path / "fake.bin"
    fake_bin.touch()
    with raises(SystemExit):
        GrizzlyArgs().parse_args(
            argv=[str(fake_bin), "adpt", "--platform", "targ"] + args
        )
    assert msg in capsys.readouterr()[-1]


@mark.skipif(system() != "Linux", reason="Only supported on Linux")
@mark.parametrize(
    "args, msg",
    [
        # test checking perf_event_paranoid
        (
            ["--rr"],
            "error: rr needs /proc/sys/kernel/perf_event_paranoid <= 1, but it is 99",
        ),
    ],
)
def test_grizzly_args_05(capsys, mocker, tmp_path, args, msg):
    """test CommonArgs.parse_args() - debugger system checks"""
    mocker.patch(
        "grizzly.args.scan_plugins",
        autospec=True,
        side_effect=(["targ"], ["adpt"]),
    )
    mocker.patch("grizzly.args.Path.read_text", autospec=True, return_value="99")
    fake_bin = tmp_path / "fake.bin"
    fake_bin.touch()
    with raises(SystemExit):
        GrizzlyArgs().parse_args(
            argv=[str(fake_bin), "adpt", "--platform", "targ"] + args
        )
    assert msg in capsys.readouterr()[-1]



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
# pylint: disable=protected-access
"""
unit tests for grizzly.Session
"""
from itertools import chain, count, repeat

from pytest import mark, raises

from sapphire import Sapphire, Served

from .adapter import Adapter
from .common.reporter import Report, Reporter
from .common.runner import RunResult
from .session import LogOutputLimiter, Session, SessionError
from .target import AssetManager, Result, Target

pytestmark = mark.usefixtures("patch_collector", "tmp_path_status_db_fuzz")


class SimpleAdapter(Adapter):
    def __init__(self, use_harness, remaining=None):
        super().__init__("simple")
        self.remaining = remaining
        self._use_harness = use_harness

    def setup(self, input_path, _server_map):
        if self._use_harness:
            self.enable_harness()
        self.fuzz["input"] = input_path

    def generate(self, testcase, _server_map):
        assert testcase.adapter_name == self.name
        testcase.input_fname = self.fuzz["input"]
        testcase.add_from_bytes(b"test", testcase.landing_page)
        if self.remaining is not None:
            assert self.remaining > 0
            self.remaining -= 1


@mark.parametrize(
    "harness, profiling, coverage, relaunch, iters, runtime",
    [
        # with harness, single iteration
        (True, False, False, 1, 1, 0),
        # with harness, 10 iterations relaunch every iteration
        (True, False, False, 1, 10, 0),
        # with harness, 10 iterations relaunch every other iteration
        (True, False, False, 2, 10, 0),
        # with harness, 10 iterations no relaunches
        (True, False, False, 10, 10, 0),
        # no harness, single iteration
        (False, False, False, 1, 1, 0),
        # no harness, 10 iterations
        (False, False, False, 1, 10, 0),
        # test enable profiling
        (True, True, False, 10, 10, 0),
        # test Session.dump_coverage()
        (True, True, True, 2, 2, 0),
        # with harness, runtime limit
        (True, False, False, 1, 0, 1),
    ],
)
def test_session_01(mocker, harness, profiling, coverage, relaunch, iters, runtime):
    """test Session with typical fuzzer Adapter"""
    mocker.patch("grizzly.common.status.time", side_effect=count(start=1.0, step=1.0))
    server = mocker.Mock(spec_set=Sapphire, port=0x1337)
    target = mocker.Mock(
        spec_set=Target,
        assets=mocker.Mock(spec_set=AssetManager),
        environ={},
        launch_timeout=30,
    )
    target.log_size.return_value = 1000
    target.monitor.launches = 1
    # avoid shutdown delay
    target.monitor.is_healthy.return_value = False
    # we can only test iter limit OR runtime limit not both
    assert bool(iters) != bool(runtime), "test is broken!"
    max_iters = iters or 1
    # calculate if the target is 'closed' based on relaunch
    type(target).closed = mocker.PropertyMock(
        side_effect=((x % relaunch == 0) for x in range(max_iters))
    )
    with Session(
        SimpleAdapter(harness),
        None,
        server,
        target,
        coverage=coverage,
        enable_profiling=profiling,
        relaunch=relaunch,
    ) as session:
        server.serve_path = lambda *a, **kv: (
            Served.ALL,
            [session.iomanager.page_name(offset=-1)],
        )
        session.run(
            [],
            10,
            input_path="file.bin",
            iteration_limit=iters,
            runtime_limit=runtime,
        )
        assert session.status.iteration == max_iters
        assert session.status.test_name == "file.bin"
        assert target.close.call_count == max_iters / relaunch
        assert target.check_result.call_count == max_iters
        assert target.handle_hang.call_count == 0
        if profiling:
            assert any(session.status.profile_entries()) == profiling
        else:
            assert not any(session.status.profile_entries())
        if coverage:
            assert target.dump_coverage.call_count == max_iters
        else:
            assert target.dump_coverage.call_count == 0


@mark.parametrize(
    "harness, relaunch, remaining",
    [
        # no harness, 1 iteration
        (False, 1, 1),
        # no harness, 5 iterations, relaunch every iteration
        (False, 1, 5),
        # harness, 1 iteration
        (True, 1, 1),
        # harness, 10 iterations, relaunch every other iteration
        (True, 2, 10),
        # harness, 10 iterations
        (True, 10, 10),
    ],
)
def test_session_02(mocker, harness, relaunch, remaining):
    """test Session with playback Adapter"""
    server = mocker.Mock(spec_set=Sapphire, port=0x1337)
    target = mocker.Mock(
        spec_set=Target,
        assets=mocker.Mock(spec_set=AssetManager),
        environ={},
        launch_timeout=30,
    )
    # calculate if the target is 'closed' based on relaunch
    type(target).closed = mocker.PropertyMock(
        side_effect=((x % relaunch == 0) for x in range(remaining))
    )
    # avoid shutdown delay
    target.monitor.is_healthy.return_value = False
    # set target.log_size to test warning code path
    target.log_size.return_value = Session.TARGET_LOG_SIZE_WARN + 1
    with Session(
        SimpleAdapter(harness, remaining=remaining),
        None,
        server,
        target,
        relaunch=relaunch,
    ) as session:
        server.serve_path = lambda *a, **kv: (
            Served.ALL,
            [session.iomanager.page_name(offset=-1)],
        )
        session.run([], 10)
        assert session.status.iteration == remaining
        assert session.status.test_name is None
        assert target.check_result.call_count == remaining
        assert target.handle_hang.call_count == 0


@mark.parametrize(
    "harness, report_size, relaunch, iters, has_sig",
    [
        # with harness, collect 1 test case
        (True, 1, 1, 1, True),
        # with harness, collect 2 test cases
        (True, 2, 2, 2, True),
        # with harness, collect 2 test cases,
        (True, 2, 3, 3, True),
        # with harness, collect 3 test cases, relaunch 1
        (True, 3, 1, 3, True),
        # without harness, collect 1 test case
        (False, 1, 1, 1, True),
        # without harness, collect 1 test case, 3 iterations
        (False, 1, 1, 3, True),
        # with harness, collect 1 test case, failed to generate FM signature
        (True, 1, 1, 1, False),
    ],
)
def test_session_03(mocker, tmp_path, harness, report_size, relaunch, iters, has_sig):
    """test Session - detecting failure"""
    adapter = SimpleAdapter(harness)
    reporter = mocker.Mock(spec_set=Reporter)
    server = mocker.Mock(spec_set=Sapphire, port=0x1337)
    target = mocker.Mock(
        spec_set=Target,
        assets=mocker.Mock(spec_set=AssetManager),
        environ={},
        launch_timeout=30,
    )
    target.monitor.launches = 1
    # avoid shutdown delay
    target.monitor.is_healthy.return_value = False
    # calculate if the target is 'closed' based on relaunch
    type(target).closed = mocker.PropertyMock(
        side_effect=((x % relaunch == 0) for x in range(iters))
    )
    # failure is on final iteration
    target.check_result.side_effect = chain(
        repeat(Result.NONE, iters - 1), (Result.FOUND,)
    )
    target.log_size.return_value = 1
    # create Report
    log_path = tmp_path / "logs"
    log_path.mkdir()
    (log_path / "log_stderr.txt").write_bytes(b"STDERR log")
    (log_path / "log_stdout.txt").write_bytes(b"STDOUT log")
    if has_sig:
        with (log_path / "log_asan_blah.txt").open("w") as log_fp:
            log_fp.write("==1==ERROR: AddressSanitizer: ")
            log_fp.write("SEGV on unknown address 0x0 (pc 0x0 bp 0x0 sp 0x0 T0)\n")
            log_fp.write("    #0 0xbad000 in foo /file1.c:123:234\n")
            log_fp.write("    #1 0x1337dd in bar /file2.c:1806:19\n")
    report = Report(log_path, "bin")
    target.create_report.return_value = report
    with Session(
        adapter, reporter, server, target, relaunch=relaunch, report_size=report_size
    ) as session:
        server.serve_path = lambda *a, **kv: (
            Served.ALL,
            [session.iomanager.page_name(offset=-1)],
        )
        session.run([], 10, input_path="file.bin", iteration_limit=iters)
        assert reporter.submit.call_count == 1
        assert len(reporter.submit.call_args[0][0]) == min(report_size, relaunch)


def test_session_04(mocker):
    """test Adapter creating invalid test case"""

    class FuzzAdapter(Adapter):
        def generate(self, _testcase, _server_map):
            pass

    server = mocker.Mock(spec_set=Sapphire, port=0x1337)
    server.serve_path.return_value = (Served.NONE, [])
    target = mocker.Mock(
        spec_set=Target,
        assets=mocker.Mock(spec_set=AssetManager),
        environ={},
        launch_timeout=30,
    )
    target.monitor.launches = 1
    with Session(FuzzAdapter("fuzz"), None, server, target) as session:
        with raises(SessionError, match="Test case is missing landing page"):
            session.run([], 10)


def test_session_05(mocker):
    """test Target not requesting landing page"""
    server = mocker.Mock(spec_set=Sapphire, port=0x1337)
    server.serve_path.return_value = (Served.TIMEOUT, [])
    target = mocker.Mock(
        spec_set=Target,
        assets=mocker.Mock(spec_set=AssetManager),
        environ={},
        launch_timeout=30,
    )
    target.monitor.launches = 1
    with Session(SimpleAdapter(False), None, server, target) as session:
        with raises(SessionError, match="Please check Adapter and Target"):
            session.run([], 10)
    assert target.handle_hang.call_count == 1


@mark.parametrize(
    "harness, report_size",
    [
        # with harness, collect 1 test case
        (True, 1),
        # with harness, collect 2 test cases
        (True, 2),
        # without harness, collect 1 test case
        (False, 1),
    ],
)
def test_session_06(mocker, harness, report_size):
    """test Session - handle Target delayed failures"""
    reporter = mocker.Mock(spec_set=Reporter)
    report = mocker.Mock(
        spec_set=Report, major="major123", minor="minor456", crash_hash="1234"
    )
    report.crash_info.createShortSignature.return_value = "[@ sig]"
    server = mocker.Mock(spec_set=Sapphire, port=0x1337)
    target = mocker.Mock(
        spec_set=Target,
        assets=mocker.Mock(spec_set=AssetManager),
        environ={},
        launch_timeout=30,
    )
    target.monitor.launches = 1
    type(target).closed = mocker.PropertyMock(side_effect=(True, False))
    target.check_result.side_effect = (Result.NONE, Result.FOUND)
    target.log_size.return_value = 1
    target.create_report.return_value = report
    with Session(
        SimpleAdapter(harness),
        reporter,
        server,
        target,
        relaunch=2,
        report_size=report_size,
    ) as session:
        server.serve_path.side_effect = (
            (Served.ALL, [session.iomanager.page_name()]),
            (Served.NONE, []),
        )
        session.run([], 10, iteration_limit=2)
        assert reporter.submit.call_count == 1
        assert len(reporter.submit.call_args[0][0]) == 1
        assert reporter.submit.call_args[0][1].major == "major123"


@mark.parametrize(
    "srv_results, target_result, ignored, results",
    [
        # delayed startup crash
        (Served.NONE, Result.FOUND, 0, 1),
        # startup hang/unresponsive
        (Served.TIMEOUT, Result.NONE, 1, 0),
    ],
)
def test_session_07(mocker, srv_results, target_result, ignored, results):
    """test Session.run() - initial test case was not served"""
    report = mocker.Mock(
        spec_set=Report, major="major123", minor="minor456", crash_hash="123"
    )
    report.crash_info.createShortSignature.return_value = "[@ sig]"
    reporter = mocker.Mock(spec_set=Reporter)
    server = mocker.Mock(spec_set=Sapphire, port=0x1337)
    target = mocker.Mock(
        spec_set=Target,
        assets=mocker.Mock(spec_set=AssetManager),
        closed=True,
        environ={},
        launch_timeout=30,
    )
    target.monitor.launches = 1
    target.check_result.side_effect = (target_result,)
    target.create_report.return_value = report
    with Session(SimpleAdapter(False), reporter, server, target) as session:
        server.serve_path.return_value = (srv_results, [])
        with raises(SessionError, match="Please check Adapter and Target"):
            session.run([], 10, iteration_limit=2)
        assert session.status.iteration == 1
        assert session.status.results.total == results
        assert session.status.ignored == ignored
        assert reporter.submit.call_count == results
        assert target.check_result.call_count == results
        assert target.handle_hang.call_count == ignored


def test_session_08(mocker):
    """test Session.run() ignoring failures"""
    result = RunResult([], 0.1, status=Result.IGNORED)
    result.attempted = True
    runner = mocker.patch("grizzly.session.Runner", autospec=True)
    runner.return_value.run.return_value = result
    runner.return_value.startup_failure = False
    adapter = mocker.Mock(spec_set=Adapter, remaining=None)
    adapter.IGNORE_UNSERVED = False
    server = mocker.Mock(spec_set=Sapphire, port=0x1337)
    target = mocker.Mock(
        spec_set=Target, assets=mocker.Mock(spec_set=AssetManager), environ={}
    )
    target.monitor.launches = 1
    with Session(adapter, None, server, target) as session:
        session.run([], 10, iteration_limit=1)
        assert runner.return_value.run.call_count == 1
        assert adapter.on_served.call_count == 1
        assert adapter.on_timeout.call_count == 0
        assert target.create_report.call_count == 0
        assert session.status.iteration == 1
        assert session.status.results.total == 0
        assert session.status.ignored == 1


def test_session_09(mocker):
    """test Session.run() report hang"""
    result = RunResult([], 60.0, status=Result.FOUND, timeout=True)
    result.attempted = True
    runner = mocker.patch("grizzly.session.Runner", autospec=True)
    runner.return_value.run.return_value = result
    runner.return_value.startup_failure = False
    adapter = mocker.Mock(spec_set=Adapter, remaining=None)
    report = mocker.Mock(spec_set=Report, major="major123", minor="minor456")
    reporter = mocker.Mock(spec_set=Reporter)
    server = mocker.Mock(spec_set=Sapphire, port=0x1337)
    target = mocker.Mock(
        spec_set=Target,
        assets=mocker.Mock(spec_set=AssetManager),
        environ={},
    )
    target.monitor.launches = 1
    target.create_report.return_value = report
    with Session(adapter, reporter, server, target) as session:
        session.run([], 10, iteration_limit=1)
        assert runner.return_value.run.call_count == 1
        assert adapter.on_served.call_count == 0
        assert adapter.on_timeout.call_count == 1
        assert target.create_report.call_count == 1
        assert reporter.submit.call_count == 1
        assert session.status.iteration == 1
        assert session.status.results.total == 1
        assert session.status.ignored == 0


@mark.parametrize(
    "harness, report_size, relaunch, iters, report_limit",
    [
        # no limit, always submit reports
        (True, 1, 1, 10, 0),
        # limit, only submit initial reports
        (True, 1, 1, 10, 5),
    ],
)
def test_session_10(mocker, harness, report_size, relaunch, iters, report_limit):
    """test Session - limit report submission"""
    adapter = SimpleAdapter(harness)
    reporter = mocker.Mock(spec_set=Reporter)
    report = mocker.Mock(spec_set=Report, major="abc", minor="def", crash_hash="123")
    report.crash_info.createShortSignature.return_value = "[@ sig]"
    server = mocker.Mock(spec_set=Sapphire, port=0x1337)
    target = mocker.Mock(
        spec_set=Target,
        assets=mocker.Mock(spec_set=AssetManager),
        environ={},
        launch_timeout=30,
    )
    target.monitor.launches = 1
    # avoid shutdown delay
    target.monitor.is_healthy.return_value = False
    type(target).closed = mocker.PropertyMock(return_value=True)
    target.check_result.return_value = Result.FOUND
    target.log_size.return_value = 1
    target.create_report.return_value = report
    with Session(
        adapter, reporter, server, target, report_limit=report_limit, relaunch=1
    ) as session:
        server.serve_path = lambda *a, **kv: (
            Served.ALL,
            [session.iomanager.page_name(offset=-1)],
        )
        session.run([], 10, input_path="file.bin", iteration_limit=iters)
        if report_limit > 0:
            assert reporter.submit.call_count == report_limit
        else:
            assert reporter.submit.call_count == iters
        assert len(reporter.submit.call_args[0][0]) == min(report_size, relaunch)
        assert reporter.submit.call_args[0][1].major == "abc"


@mark.parametrize(
    "harness, iters, result_limit, results",
    [
        # hit result limit
        (True, 0, 1, 1),
        # don't hit result limit (one result)
        (True, 10, 5, 1),
        # don't hit result limit (no results)
        (True, 2, 1, 0),
    ],
)
def test_session_12(mocker, harness, iters, result_limit, results):
    """test Session - limit results"""
    adapter = SimpleAdapter(harness)
    reporter = mocker.Mock(spec_set=Reporter)
    report = mocker.Mock(spec_set=Report, major="abc", minor="def", crash_hash="123")
    report.crash_info.createShortSignature.return_value = "[@ sig]"
    server = mocker.Mock(spec_set=Sapphire, port=0x1337)
    target = mocker.Mock(
        spec_set=Target,
        assets=mocker.Mock(spec_set=AssetManager),
        environ={},
        launch_timeout=30,
    )
    target.monitor.launches = 1
    # avoid shutdown delay
    target.monitor.is_healthy.return_value = False
    type(target).closed = mocker.PropertyMock(return_value=True)
    target.check_result.side_effect = chain(
        repeat(Result.FOUND, results), repeat(Result.NONE, iters - results)
    )
    target.log_size.return_value = 1
    target.create_report.return_value = report
    with Session(adapter, reporter, server, target, relaunch=1) as session:
        server.serve_path = lambda *a, **kv: (
            Served.ALL,
            [session.iomanager.page_name(offset=-1)],
        )
        session.run(
            [], 10, input_path="a.bin", iteration_limit=iters, result_limit=result_limit
        )
    if results >= result_limit > iters:
        # limited by result_limit
        assert session.status.iteration == result_limit
    else:
        assert session.status.iteration == iters


def test_log_output_limiter_01(mocker):
    """test LogOutputLimiter.ready() not ready"""
    fake_time = mocker.patch("grizzly.session.time", autospec=True)
    fake_time.return_value = 1.0
    lol = LogOutputLimiter(delay=10, delta_multiplier=2)
    assert lol._delay == 10
    assert lol._iterations == 1
    assert lol._launches == 1
    assert lol._multiplier == 2
    assert lol._time == 1.0
    assert not lol._verbose
    fake_time.return_value = 1.1
    assert not lol.ready(0, 0)
    assert lol._iterations == 1
    assert lol._launches == 1
    assert lol._time == 1.0
    lol._verbose = True
    assert lol.ready(0, 0)


def test_log_output_limiter_02(mocker):
    """test LogOutputLimiter.ready() due to iterations"""
    fake_time = mocker.patch("grizzly.session.time", autospec=True)
    fake_time.return_value = 1.0
    lol = LogOutputLimiter(delay=10, delta_multiplier=2)
    fake_time.return_value = 1.1
    lol._launches = 2
    assert lol.ready(1, 1)
    assert lol._iterations == 2
    assert lol._launches == 2
    assert lol._time == 1.1


def test_log_output_limiter_03(mocker):
    """test LogOutputLimiter.ready() due to launches"""
    fake_time = mocker.patch("grizzly.session.time", autospec=True)
    fake_time.return_value = 1.0
    lol = LogOutputLimiter(delay=10, delta_multiplier=2)
    lol._iterations = 4
    assert lol.ready(3, 1)
    assert lol._launches == 2
    assert lol._iterations == 4
    assert lol._time == 1.0


def test_log_output_limiter_04(mocker):
    """test LogOutputLimiter.ready() due to time"""
    fake_time = mocker.patch("grizzly.session.time", autospec=True)
    fake_time.return_value = 1.0
    lol = LogOutputLimiter(delay=1, delta_multiplier=2)
    lol._iterations = 4
    fake_time.return_value = 2.0
    assert lol.ready(3, 0)
    assert lol._iterations == 4
    assert lol._launches == 1
    assert lol._time == 2.0



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
from argparse import SUPPRESS
from pathlib import Path

from ..args import CommonArgs

LOCAL_INPUT_HELP = """Accepted input:
A directory containing testcase data;
A directory with one or more subdirectories containing testcase data;
A zip archive containing testcase data or subdirectories containing testcase data;
or a single file to be used as a testcase.
NOTE: When using a directory it must contain a 'test_info.json' file."""


class ReplayCommonArgs(CommonArgs):
    def __init__(self):
        super().__init__()
        self.parser.set_defaults(logs=None)

        replay_args = self.parser.add_argument_group("Replay Arguments")
        replay_args.add_argument(
            "--any-crash",
            action="store_true",
            help="Any crash is interesting, not only crashes which match the original"
            " signature.",
        )
        replay_args.add_argument(
            "--idle-delay",
            type=int,
            default=30,
            help="Number of seconds to wait before polling for idle"
            " (default: %(default)s)",
        )
        replay_args.add_argument(
            "--idle-threshold",
            type=int,
            default=0,
            help="CPU usage threshold to mark the process as idle (default: disabled)",
        )
        replay_args.add_argument(
            "--min-crashes",
            type=int,
            default=1,
            help="Require the testcase to crash n times before accepting the result."
            " Helpful for intermittent testcases (default: %(default)sx)",
        )
        # hidden argument to add original crash ID as metadata when reported
        replay_args.add_argument("--original-crash-id", type=int, help=SUPPRESS)
        replay_args.add_argument(
            "--repeat",
            type=int,
            default=1,
            help="Run the testcase n times."
            " Helpful for intermittent testcases (default: %(default)sx)",
        )
        replay_args.add_argument(
            "--sig", type=Path, help="Signature (JSON) file to match."
        )
        replay_args.add_argument(
            "--test-index",
            type=int,
            nargs="+",
            help="Select a testcase to run when multiple testcases are loaded. "
            "Testscases are ordered oldest to newest. Indexing is 0 based. "
            "0 == Oldest, n-1 == Newest (default: run all testcases)",
        )

        self.launcher_grp.add_argument(
            "--post-launch-delay",
            type=int,
            default=None,
            help="Number of seconds to wait after launching browser before "
            " running the test. This provides an opportunity to connect debuggers. ",
        )

        self.reporter_grp.add_argument(
            "--include-test",
            action="store_true",
            help="Include the testcase when reporting results.",
        )

    def sanity_check(self, args):
        super().sanity_check(args)

        if args.any_crash and args.sig is not None:
            self.parser.error("signature is ignored when running with --any-crash")

        if args.idle_threshold and args.idle_delay <= 0:
            self.parser.error("--idle-delay value must be positive")

        if args.logs is None and (args.pernosco or args.rr):
            self.parser.error("--logs must be set when using rr")

        if args.min_crashes < 1:
            self.parser.error("--min-crashes value must be positive")

        if args.post_launch_delay is not None and args.post_launch_delay < 0:
            self.parser.error("--post-launch-delay value must be positive")

        if args.repeat < 1:
            self.parser.error("--repeat value must be positive")

        if args.sig and not args.sig.is_file():
            self.parser.error(f"signature file not found: '{args.sig}'")


class ReplayArgs(ReplayCommonArgs):
    def __init__(self):
        super().__init__()
        self.parser.add_argument("input", type=Path, help=LOCAL_INPUT_HELP)

    def sanity_check(self, args):
        super().sanity_check(args)

        if not args.input.exists():
            self.parser.error(f"'{args.input}' does not exist")


class ReplayFuzzManagerIDArgs(ReplayCommonArgs):
    def __init__(self):
        """Initialize argument parser."""
        super().__init__()
        self.parser.add_argument("input", type=int, help="FuzzManager ID to replay")


class ReplayFuzzManagerIDQualityArgs(ReplayFuzzManagerIDArgs):
    def __init__(self):
        """Initialize argument parser."""
        super().__init__()
        self.parser.add_argument(
            "--quality", type=int, help="Only try crashes with a given quality value"
        )

    def sanity_check(self, args):
        super().sanity_check(args)

        if args.quality is not None and args.quality < 0:
            self.parser.error("'--quality' value cannot be negative")



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
from logging import getLogger

from pkg_resources import iter_entry_points

__all__ = ("load", "scan", "PluginLoadError")


LOG = getLogger(__name__)


class PluginLoadError(Exception):
    """Raised if loading a plug-in fails"""


def load(name, group, base_type):
    """Load a plug-in.

    Args:
        name (str): Name of entry point to load.
        group (str): Group containing entry point.
        base_type (type): Used to validate loaded objects.

    Returns:
        *: Python object.
    """
    assert isinstance(base_type, type)
    for entry in iter_entry_points(group):
        if entry.name == name:
            LOG.debug("loading %r (%s)", name, base_type.__name__)
            plugin = entry.load()
            break
    else:
        raise PluginLoadError(f"{name!r} not found in {group!r}")
    if not issubclass(plugin, base_type):
        raise PluginLoadError(f"{name!r} doesn't inherit from {base_type.__name__}")
    return plugin


def scan(group):
    """Scan for installed plug-ins.

    Args:
        group (str): Entry point group to scan.

    Returns:
        list: Names of installed entry points.
    """
    found = []
    LOG.debug("scanning %r", group)
    for entry in iter_entry_points(group):
        if entry.name in found:
            # not sure if this can even happen
            raise PluginLoadError(f"Duplicate entry {entry.name!r} in {group!r}")
        found.append(entry.name)
    return found


def scan_target_assets():
    """Scan targets and load list of supported assets (minimal sanity checking).

    Args:
        None

    Returns:
        dict: Name of target and list of supported assets.
    """
    assets = {}
    for entry in iter_entry_points("grizzly_targets"):
        assets[entry.name] = entry.load().SUPPORTED_ASSETS
    return assets



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
"""Tests for interface for getting Crash and Bucket data from CrashManager API"""
import json

from pytest import mark, raises

from .fuzzmanager import Bucket, CrashEntry, load_fm_data


def test_bucket_1(mocker):
    """bucket getattr uses data from get"""
    coll = mocker.patch("grizzly.common.fuzzmanager.Collector", autospec=True)
    coll.return_value.get.return_value.json.return_value = {"testcase": "data"}
    coll.return_value.serverProtocol = "http"
    coll.return_value.serverPort = 123
    coll.return_value.serverHost = "allizom.org"
    bucket = Bucket(123)
    assert coll.return_value.get.call_count == 0
    assert bucket.testcase == "data"
    with raises(AttributeError):
        getattr(bucket, "other")
    assert coll.return_value.get.call_count == 1


def test_bucket_2(mocker):
    """bucket setattr raises"""
    coll = mocker.patch("grizzly.common.fuzzmanager.Collector", autospec=True)
    coll.return_value.serverProtocol = "http"
    coll.return_value.serverPort = 123
    coll.return_value.serverHost = "allizom.org"
    bucket = Bucket(123)
    with raises(AttributeError):
        bucket.other = "data"
    assert coll.return_value.get.call_count == 0


def test_bucket_3(mocker):
    """bucket iter_crashes flattens across pages"""
    coll = mocker.patch("grizzly.common.fuzzmanager.Collector", autospec=True)
    coll.return_value.serverProtocol = "http"
    coll.return_value.serverPort = 123
    coll.return_value.serverHost = "allizom.org"
    coll.return_value.get.return_value.json.side_effect = [
        {
            "count": 2,
            "next": "url",
            "results": [
                {"id": 234, "testcase": "test1"},
                {"id": 345, "testcase": None},
            ],
        },
        {
            "count": 1,
            "next": None,
            "results": [
                {"id": 456, "testcase": "test2"},
            ],
        },
    ]
    bucket = Bucket(123)
    assert coll.return_value.get.call_count == 0
    crashes = list(bucket.iter_crashes(quality_filter=5))
    assert coll.return_value.get.call_count == 2
    assert coll.return_value.get.call_args_list[0][1]["params"]["include_raw"] == "0"
    assert json.loads(
        coll.return_value.get.call_args_list[0][1]["params"]["query"]
    ) == {
        "op": "AND",
        "bucket": 123,
        "testcase__quality": 5,
    }
    assert len(crashes) == 2
    assert crashes[0].crash_id == 234
    assert crashes[1].crash_id == 456


def test_bucket_4(mocker):
    """bucket signature_path writes and returns sig json and metadata"""
    coll = mocker.patch("grizzly.common.fuzzmanager.Collector", autospec=True)
    coll.return_value.serverProtocol = "http"
    coll.return_value.serverPort = 123
    coll.return_value.serverHost = "allizom.org"
    coll.return_value.get.return_value.json.return_value = {
        "signature": "sigdata",
        "size": 10,
        "frequent": True,
        "shortDescription": "sig desc",
        "best_quality": 0,
    }
    with Bucket(123) as bucket:
        assert coll.return_value.get.call_count == 0
        sig_path = bucket.signature_path()
        assert sig_path.is_file()
        assert sig_path.with_suffix(".metadata").is_file()
        assert sig_path.read_text() == "sigdata"
        assert json.loads(sig_path.with_suffix(".metadata").read_text()) == {
            "size": 10,
            "frequent": True,
            "shortDescription": "sig desc",
            "testcase__quality": 0,
        }
        assert coll.return_value.get.call_count == 1
        # second call returns same path
        assert bucket.signature_path() == sig_path
    assert coll.return_value.get.call_count == 1


def test_crash_1(mocker):
    """crash getattr uses data from get"""
    coll = mocker.patch("grizzly.common.fuzzmanager.Collector", autospec=True)
    coll.return_value.get.return_value.json.return_value = {"testcase": "data"}
    coll.return_value.serverProtocol = "http"
    coll.return_value.serverPort = 123
    coll.return_value.serverHost = "allizom.org"
    crash = CrashEntry(123)
    assert coll.return_value.get.call_count == 0
    assert crash.testcase == "data"
    with raises(AttributeError):
        getattr(crash, "other")
    assert coll.return_value.get.call_count == 1
    assert coll.return_value.get.call_args[1]["params"] == {"include_raw": "0"}

    # crash getattr for raw field re-gets
    coll.return_value.get.return_value.json.return_value = {"rawStderr": "stderr"}
    assert crash.rawStderr == "stderr"
    assert coll.return_value.get.call_count == 2
    assert coll.return_value.get.call_args[1]["params"] == {"include_raw": "1"}


def test_crash_2(mocker):
    """crash setattr raises except testcase_quality"""
    coll = mocker.patch("grizzly.common.fuzzmanager.Collector", autospec=True)
    coll.return_value.get.return_value.json.return_value = {"testcase": "data"}
    coll.return_value.serverProtocol = "http"
    coll.return_value.serverPort = 123
    coll.return_value.serverHost = "allizom.org"
    crash = CrashEntry(123)

    # crash setattr raises for other field
    with raises(AttributeError):
        crash.other = "data"
    assert coll.return_value.get.call_count == 0

    # crash setattr for testcase_quality works and updates data if set
    assert coll.return_value.patch.call_count == 0
    crash.testcase_quality = 5
    assert coll.return_value.get.call_count == 0
    assert coll.return_value.patch.call_count == 1
    with raises(AttributeError):
        getattr(crash, "testcase_quality")
    assert coll.return_value.get.call_count == 1
    getattr(crash, "testcase")
    assert coll.return_value.get.call_count == 1
    crash.testcase_quality = 10
    assert coll.return_value.patch.call_count == 2
    assert crash.testcase_quality == 10
    assert coll.return_value.get.call_count == 1


def test_crash_3(mocker):
    """crash testcase_path writes and returns testcase zip"""
    coll = mocker.patch("grizzly.common.fuzzmanager.Collector", autospec=True)
    coll.return_value.serverProtocol = "http"
    coll.return_value.serverPort = 123
    coll.return_value.serverHost = "allizom.org"
    coll.return_value.get.return_value.json.return_value = {
        "id": 234,
        "testcase": "test.bz2",
    }
    with CrashEntry(234) as crash:
        assert crash.testcase == "test.bz2"  # pre-load data dict so I can re-patch get
        coll.return_value.get.return_value = mocker.Mock(
            content=b"\x01\x02\x03",
            headers={"content-disposition"},
        )
        assert coll.return_value.get.call_count == 1
        tc_path = crash.testcase_path()
        assert tc_path.is_file()
        assert tc_path.suffix == ".bz2"
        assert tc_path.read_bytes() == b"\x01\x02\x03"
        assert coll.return_value.get.call_count == 2
        # second call returns same path
        assert crash.testcase_path() == tc_path
    assert coll.return_value.get.call_count == 2


@mark.parametrize(
    "bucket_id, load_bucket",
    [
        # Nothing to load, don't try
        (None, False),
        # Nothing to load, try
        (None, True),
        # Bucket exists, don't load it
        (111, False),
        # Bucket exists, load it
        (111, True),
    ],
)
def test_load_fm_data_1(mocker, bucket_id, load_bucket):
    """test load_fm_data()"""
    coll = mocker.patch("grizzly.common.fuzzmanager.Collector", autospec=True)
    coll.return_value.serverProtocol = "http"
    coll.return_value.serverPort = 123
    coll.return_value.serverHost = "allizom.org"
    coll.return_value.get.return_value = mocker.Mock(
        content=b"\x01\x02\x03",
        headers={"content-disposition"},
    )
    coll.return_value.get.return_value.json.return_value = {"bucket": bucket_id}

    with load_fm_data(123, load_bucket) as (crash, bucket):
        assert isinstance(crash, CrashEntry)
        if load_bucket and bucket_id:
            assert isinstance(bucket, Bucket)
        else:
            assert bucket is None



"""
Sapphire unit tests
"""
# pylint: disable=protected-access

import hashlib
import os
import random
import socket
import threading
from platform import system
from urllib.parse import quote, urlparse

from pytest import mark, raises

from .core import Sapphire
from .job import Served
from .server_map import ServerMap
from .worker import Worker


class _TestFile:
    def __init__(self, url, url_prefix=None):
        self.code = None
        self.content_type = None
        self.custom_request = None
        if url_prefix:
            self.file = "".join((url_prefix, url))
        else:
            self.file = url
        self.len_org = 0  # original file length
        self.len_srv = 0  # served file length
        self.lock = threading.Lock()
        self.md5_org = None
        self.md5_srv = None
        self.requested = 0  # number of time file was requested
        url = urlparse(self.file.replace("\\", "/"))
        self.url = (
            "?".join((quote(url.path), url.query)) if url.query else quote(url.path)
        )


def _create_test(fname, path, data=b"Test!", calc_hash=False, url_prefix=None):
    test = _TestFile(fname, url_prefix=url_prefix)
    with (path / fname).open("w+b") as out_fp:
        out_fp.write(data)
        test.len_org = out_fp.tell()
        if calc_hash:
            out_fp.seek(0)
            test.md5_org = hashlib.md5(out_fp.read()).hexdigest()
    return test


def test_sapphire_00(client, tmp_path):
    """test requesting a single file"""
    with Sapphire(timeout=10) as serv:
        assert serv.timeout == 10
        test = _create_test("test_case.html", tmp_path)
        client.launch("127.0.0.1", serv.port, [test])
        assert serv.serve_path(tmp_path)[0] == Served.ALL
    assert client.wait(timeout=10)
    assert test.code == 200
    assert test.len_srv == test.len_org


def test_sapphire_01(client, tmp_path):
    """test requesting multiple files (test cleanup code)"""
    to_serve = [
        _create_test(f"test_{i}.html", tmp_path, data=os.urandom(5), calc_hash=True)
        for i in range(100)
    ]
    with Sapphire(timeout=30) as serv:
        client.launch("127.0.0.1", serv.port, to_serve)
        status, files_served = serv.serve_path(tmp_path)
    assert status == Served.ALL
    assert len(to_serve) == len(files_served)
    assert client.wait(timeout=10)
    for t_file in to_serve:
        assert t_file.code == 200
        assert t_file.len_srv == t_file.len_org
        assert t_file.md5_srv == t_file.md5_org


def test_sapphire_02(client, tmp_path):
    """test serving optional file"""
    files_to_serve = [_create_test(f"test_{i}.html", tmp_path) for i in range(3)]
    optional = [files_to_serve[0].file]
    with Sapphire(timeout=10) as serv:
        client.launch("127.0.0.1", serv.port, files_to_serve, in_order=True)
        status, served_list = serv.serve_path(tmp_path, optional_files=optional)
    assert status == Served.ALL
    assert len(files_to_serve) == len(served_list)
    assert client.wait(timeout=10)
    for t_file in files_to_serve:
        assert t_file.code == 200
        assert t_file.len_srv == t_file.len_org


def test_sapphire_03(client, tmp_path):
    """test skipping optional file"""
    files_to_serve = [_create_test(f"test_{i}.html", tmp_path) for i in range(3)]
    optional = [files_to_serve[0].file]
    with Sapphire(timeout=10) as serv:
        client.launch("127.0.0.1", serv.port, files_to_serve[1:])
        status, served_list = serv.serve_path(tmp_path, optional_files=optional)
    assert status == Served.ALL
    assert len(served_list) == len(files_to_serve) - 1
    assert client.wait(timeout=10)
    assert files_to_serve[0].code is None
    assert files_to_serve[0].len_srv == 0
    for t_file in files_to_serve[1:]:
        assert t_file.code == 200
        assert t_file.len_srv == t_file.len_org


def test_sapphire_04(client, tmp_path):
    """test requesting invalid file (404)"""
    files_to_serve = [
        _TestFile("does_not_exist.html"),
        _create_test("test_case.html", tmp_path),
    ]
    with Sapphire(timeout=10) as serv:
        client.launch("127.0.0.1", serv.port, files_to_serve, in_order=True)
        assert serv.serve_path(tmp_path)[0] == Served.ALL
    assert client.wait(timeout=10)
    assert "does_not_exist.html" in files_to_serve[0].file
    assert files_to_serve[0].code == 404
    assert files_to_serve[1].code == 200


def test_sapphire_05(client, tmp_path):
    """test requesting a file outside of the server root (403)"""
    root_dir = tmp_path / "root"
    root_dir.mkdir()
    files_to_serve = [
        # add invalid file
        _TestFile(os.path.abspath(__file__)),
        # add file in parent of root_dir
        _create_test("no_access.html", tmp_path, data=b"no_access", url_prefix="../"),
        # add valid test
        _create_test("test_case.html", root_dir),
    ]
    assert (tmp_path / "no_access.html").is_file()
    with Sapphire(timeout=10) as serv:
        client.launch("127.0.0.1", serv.port, files_to_serve, in_order=True)
        status, files_served = serv.serve_path(root_dir)
    assert status == Served.ALL
    assert len(files_served) == 1
    assert client.wait(timeout=10)
    assert os.path.basename(__file__) in files_to_serve[0].file
    assert files_to_serve[0].code == 403
    assert "no_access.html" in files_to_serve[1].file
    assert files_to_serve[1].code == 403
    assert files_to_serve[2].code == 200


def test_sapphire_06(client, tmp_path):
    """test serving no files... this should never happen but..."""
    with Sapphire(timeout=10) as serv:
        client.launch("127.0.0.1", serv.port, [])
        status, files_served = serv.serve_path(tmp_path)
    assert status == Served.NONE
    assert not files_served


def test_sapphire_07(tmp_path):
    """test timeout of the server"""
    with Sapphire(timeout=60) as serv:
        assert serv.timeout == 60  # verify default
        serv.timeout = 0  # disable timeout
        assert serv.timeout == 0
        serv.timeout = 0.01
        assert serv.timeout == 0.01
        _create_test("test_case.html", tmp_path)
        status, files_served = serv.serve_path(tmp_path)
    assert status == Served.TIMEOUT
    assert not files_served


def test_sapphire_08(client, tmp_path):
    """test only serving some files (Served.REQUEST)"""
    cb_status = {"count": 0}

    def is_running():
        cb_status["count"] += 1
        return cb_status["count"] < 3  # return false after 2nd call

    files_to_serve = [_create_test(f"test_{i}.html", tmp_path) for i in range(3)]
    with Sapphire() as serv:
        client.launch("127.0.0.1", serv.port, files_to_serve[1:])
        status, files_served = serv.serve_path(tmp_path, continue_cb=is_running)
    assert status == Served.REQUEST
    assert len(files_served) < len(files_to_serve)


def test_sapphire_09(client, tmp_path):
    """test serving interesting sized files"""
    tests = [
        {"size": Worker.DEFAULT_TX_SIZE, "name": "even.html"},
        {"size": Worker.DEFAULT_TX_SIZE - 1, "name": "minus_one.html"},
        {"size": Worker.DEFAULT_TX_SIZE + 1, "name": "plus_one.html"},
        {"size": Worker.DEFAULT_TX_SIZE * 2, "name": "double.html"},
        {"size": 1, "name": "one.html"},
        {"size": 0, "name": "zero.html"},
    ]
    for test in tests:
        test["file"] = _TestFile(test["name"])
        t_data = "".join(random.choices("ABCD1234", k=test["size"])).encode("ascii")
        (tmp_path / test["file"].file).write_bytes(t_data)
        test["file"].md5_org = hashlib.md5(t_data).hexdigest()
    with Sapphire(timeout=10) as serv:
        client.launch("127.0.0.1", serv.port, [test["file"] for test in tests])
        status, served_list = serv.serve_path(tmp_path)
    assert status == Served.ALL
    assert len(served_list) == len(tests)
    assert client.wait(timeout=10)
    for test in tests:
        assert test["file"].code == 200
        assert test["file"].len_srv == test["size"]
        assert test["file"].md5_srv == test["file"].md5_org


def test_sapphire_10(client, tmp_path):
    """test serving a large (100MB) file"""
    t_file = _TestFile("test_case.html")
    data_hash = hashlib.md5()
    with (tmp_path / t_file.file).open("wb") as test_fp:
        # write 100MB of 'A'
        data = b"A" * (100 * 1024)  # 100KB of 'A'
        for _ in range(1024):
            test_fp.write(data)
            data_hash.update(data)
    t_file.md5_org = data_hash.hexdigest()
    with Sapphire(timeout=10) as serv:
        client.launch("127.0.0.1", serv.port, [t_file])
        assert serv.serve_path(tmp_path)[0] == Served.ALL
    assert client.wait(timeout=10)
    assert t_file.code == 200
    assert t_file.len_srv == (100 * 1024 * 1024)
    assert t_file.md5_srv == t_file.md5_org


def test_sapphire_11(client, tmp_path):
    """test serving a binary file"""
    t_file = _create_test(
        "test_case.html", tmp_path, data=os.urandom(512), calc_hash=True
    )
    with Sapphire(timeout=10) as serv:
        client.launch("127.0.0.1", serv.port, [t_file])
        assert serv.serve_path(tmp_path)[0] == Served.ALL
    assert client.wait(timeout=10)
    assert t_file.code == 200
    assert t_file.len_srv == t_file.len_org
    assert t_file.md5_srv == t_file.md5_org


def test_sapphire_12():
    """test requested port is used"""
    test_port = 0x1337
    with Sapphire(port=test_port, timeout=1) as serv:
        assert test_port == serv.port


def test_sapphire_13(client, tmp_path):
    """test serving multiple content types"""
    files_to_serve = [
        _create_test("test_case.html", tmp_path),
        # create binary file without an extension
        _create_test("test_case", tmp_path, data=os.urandom(5)),
    ]
    with Sapphire(timeout=10) as serv:
        client.launch("127.0.0.1", serv.port, files_to_serve)
        assert serv.serve_path(tmp_path)[0] == Served.ALL
    assert client.wait(timeout=10)
    content_types = set()
    for test in files_to_serve:
        assert test.code == 200
        assert test.len_srv == test.len_org
        file_ext = os.path.splitext(test.file)[-1]
        content_type = {".html": "text/html"}.get(file_ext, "application/octet-stream")
        content_types.add(content_type)
        assert test.content_type == content_type
    assert len(content_types) == 2


def test_sapphire_14(tmp_path):
    """test callback"""
    cb_status = {"count": 0}

    def _test_callback():
        cb_status["count"] += 1
        # return true on first call
        return cb_status["count"] < 2

    with Sapphire(timeout=10) as serv:
        _create_test("test_case.html", tmp_path)
        assert serv.serve_path(tmp_path, continue_cb=_test_callback)[0] == Served.NONE
    assert cb_status["count"] == 2


def test_sapphire_15(client, tmp_path):
    """test calling serve_path multiple times"""
    with Sapphire(timeout=10) as serv:
        for i in range(3):
            test = _create_test(f"test_case_{i}.html", tmp_path)
            client.launch("127.0.0.1", serv.port, [test])
            assert serv.serve_path(tmp_path)[0] == Served.ALL
            assert client.wait(timeout=10)
            client.close()
            assert test.code == 200
            assert test.len_srv == test.len_org
            (tmp_path / test.file).unlink()


def test_sapphire_16(client, tmp_path):
    """test non required mapped redirects"""
    smap = ServerMap()
    smap.set_redirect("test_url", "blah", required=False)
    test = _create_test("test_case.html", tmp_path)
    with Sapphire(timeout=10) as serv:
        client.launch("127.0.0.1", serv.port, [test])
        assert serv.serve_path(tmp_path, server_map=smap)[0] == Served.ALL
    assert client.wait(timeout=10)
    assert test.code == 200
    assert test.len_srv == test.len_org


@mark.parametrize(
    "path, query",
    [
        # simple path
        ("test.html", None),
        # simple path with query
        ("test.html", "foo=1&bar=2"),
        # non-alphanumeric chars (valid characters to use on filesystem)
        ("!@$%^&(_+-=[]),'~`{}", None),
        # extended ascii chars
        ("€d’é-ñÿ", None),
    ],
)
def test_sapphire_17(client, tmp_path, path, query):
    """test required mapped redirects"""
    smap = ServerMap()
    with Sapphire(timeout=10) as serv:
        # target will be requested indirectly via the redirect
        target = _create_test(path, tmp_path, data=b"Redirect DATA!")
        request_path = "redirect" if query is None else "?".join(("redirect", query))
        redirect = _TestFile(request_path)
        # point "redirect" at target
        smap.set_redirect("redirect", target.file, required=True)
        client.launch("127.0.0.1", serv.port, [redirect])
        status, served = serv.serve_path(tmp_path, server_map=smap)
    assert status == Served.ALL
    assert len(served) == 1
    assert client.wait(timeout=10)
    assert redirect.code == 200
    assert redirect.len_srv == target.len_org


def test_sapphire_18(client, tmp_path):
    """test include directories and permissions"""
    inc1_path = tmp_path / "inc1"
    inc2_path = tmp_path / "inc2"
    root_path = tmp_path / "root"
    inc1_path.mkdir()
    inc2_path.mkdir()
    root_path.mkdir()
    files_to_serve = []
    smap = ServerMap()
    with Sapphire(timeout=10) as serv:
        # add files to inc dirs
        inc1 = _create_test("included_file1.html", inc1_path, data=b"blah....1")
        files_to_serve.append(inc1)
        # add a nested dir
        nest_path = inc1_path / "nested"
        nest_path.mkdir()
        # add file in a nested dir in inc1
        nest = _create_test(
            "nested_file.html", nest_path, data=b"blah... .nested", url_prefix="nested/"
        )
        assert nest_path / "nested_file.html"
        files_to_serve.append(nest)
        # test 404 in nested dir in inc1
        nest_404 = _TestFile("nested/nested_file_404.html")
        files_to_serve.append(nest_404)
        # test path mounted somewhere other than /
        inc2 = _create_test(
            "included_file2.html", inc2_path, data=b"blah....2", url_prefix="inc_test/"
        )
        files_to_serve.append(inc2)
        # test 404 in include dir
        inc404 = _TestFile("inc_test/included_file_404.html")
        assert not (nest_path / "included_file_404.html").is_file()
        files_to_serve.append(inc404)
        # test 403
        inc403 = _create_test(
            "no_access.html", tmp_path, data=b"no_access", url_prefix="inc_test/../"
        )
        assert (tmp_path / "no_access.html").is_file()
        files_to_serve.append(inc403)
        # test file (used to keep sever job alive)
        test = _create_test("test_case.html", root_path)
        files_to_serve.append(test)
        # add include paths
        smap.set_include("/", str(inc1_path))  # mount at '/'
        smap.set_include("inc_test", str(inc2_path))  # mount at '/inc_test'
        client.launch("127.0.0.1", serv.port, files_to_serve, in_order=True)
        status, files_served = serv.serve_path(root_path, server_map=smap)
    assert status == Served.ALL
    assert "test_case.html" in files_served
    assert (inc1_path / "included_file1.html").as_posix() in files_served
    assert (inc2_path / "included_file2.html").as_posix() in files_served
    assert (nest_path / "nested_file.html").as_posix() in files_served
    assert client.wait(timeout=10)
    assert inc1.code == 200
    assert inc2.code == 200
    assert nest.code == 200
    assert test.code == 200
    assert nest_404.code == 404
    assert inc404.code == 404
    assert inc403.code == 403


@mark.parametrize(
    "query, required",
    [
        # dynamic response not required
        (None, False),
        # dynamic response is required
        (None, True),
        # dynamic response with query string
        ("test123", True),
        # dynamic response with empty query string
        ("", True),
    ],
)
def test_sapphire_19(client, tmp_path, query, required):
    """test dynamic response"""
    _data = b"dynamic response -- TEST DATA!"
    # build request
    path = "dyn_test"
    if query is not None:
        request = "?".join([path, query])
    else:
        request = path

    # setup custom callback
    def dr_callback(data):
        if query:
            assert data == query
        else:
            assert data == ""
        return _data

    smap = ServerMap()
    smap.set_dynamic_response(
        path, dr_callback, mime_type="text/plain", required=required
    )
    # create files
    test_dr = _TestFile(request)
    test_dr.len_org = len(_data)
    test_dr.md5_org = hashlib.md5(_data).hexdigest()
    test = _create_test("test_case.html", tmp_path)
    if required:
        optional = [test.file]
        files = [test, test_dr]
    else:
        optional = [test_dr.file]
        files = [test_dr, test]
    # test request
    with Sapphire(timeout=10) as serv:
        client.launch("127.0.0.1", serv.port, files, in_order=True)
        assert (
            serv.serve_path(tmp_path, optional_files=optional, server_map=smap)[0]
            == Served.ALL
        )
    assert client.wait(timeout=10)
    if not required:
        assert test.code == 200
        assert test.len_srv == test.len_org
    assert test_dr.code == 200
    assert test_dr.len_srv == test_dr.len_org
    assert test_dr.md5_srv == test_dr.md5_org


def test_sapphire_20(client_factory, tmp_path):
    """test pending_files == 0 in worker thread"""
    client_defer = client_factory(rx_size=2)
    # server should shutdown while this file is being served
    test_defer = _create_test("defer_test.html", tmp_path)
    optional = [test_defer.file]
    test = _create_test("test_case.html", tmp_path, data=b"112233")
    with Sapphire(timeout=10) as serv:
        # this test needs to wait just long enough to have the required file served
        # but not too long or the connection will be closed by the server
        client_defer.launch(
            "127.0.0.1", serv.port, [test_defer], delay=0.1, indicate_failure=True
        )
        client = client_factory(rx_size=2)
        client.launch("127.0.0.1", serv.port, [test], throttle=0.1)
        assert serv.serve_path(tmp_path, optional_files=optional)[0] == Served.ALL
    assert client_defer.wait(timeout=10)
    assert client.wait(timeout=10)
    assert test.code == 200
    assert test_defer.code == 0


def test_sapphire_21(client, tmp_path):
    """test handling an invalid request"""
    bad_test = _TestFile("bad.html")
    bad_test.custom_request = b"a bad request...0+%\xef\xb7\xba\r\n"
    optional = [bad_test.file]
    test = _create_test("test_case.html", tmp_path)
    with Sapphire(timeout=10) as serv:
        client.launch("127.0.0.1", serv.port, [bad_test, test], in_order=True)
        assert serv.serve_path(tmp_path, optional_files=optional)[0] == Served.ALL
    assert client.wait(timeout=10)
    assert test.code == 200
    assert bad_test.code == 400


def test_sapphire_22(client, tmp_path):
    """test handling an empty request"""
    bad_test = _TestFile("bad.html")
    bad_test.custom_request = b""
    optional = [bad_test.file]
    test = _create_test("test_case.html", tmp_path)
    with Sapphire(timeout=10) as serv:
        client.launch(
            "127.0.0.1",
            serv.port,
            [bad_test, test],
            indicate_failure=True,
            in_order=True,
        )
        assert serv.serve_path(tmp_path, optional_files=optional)[0] == Served.ALL
    assert client.wait(timeout=10)
    assert test.code == 200
    assert bad_test.code == 0


def test_sapphire_23(client_factory, tmp_path):
    """test requesting multiple files via multiple connections"""
    to_serve = [
        _create_test(f"test_{i:0>3d}.html", tmp_path, data=b"AAAA") for i in range(2)
    ]
    max_workers = 20
    with Sapphire(max_workers=max_workers, timeout=60) as serv:
        clients = []
        try:
            for _ in range(max_workers):  # number of clients to spawn
                clients.append(client_factory(rx_size=1))
            for client in clients:
                client.launch(
                    "127.0.0.1", serv.port, to_serve, in_order=True, throttle=0.05
                )
            status, files_served = serv.serve_path(tmp_path)
            # call serv.close() instead of waiting for the clients to timeout
            serv.close()
        finally:
            for client in clients:
                assert client.wait(timeout=10)
                client.close()
    assert status == Served.ALL
    assert len(to_serve) == len(files_served)
    for t_file in to_serve:
        assert t_file.code == 200
        assert t_file.len_srv == t_file.len_org


def test_sapphire_24(client_factory, tmp_path):
    """test all request types via multiple connections"""

    def _dyn_test_cb(_):
        return b"A" if random.getrandbits(1) else b"AA"

    smap = ServerMap()
    with Sapphire(max_workers=10, timeout=60) as serv:
        to_serve = []
        for i in range(50):
            # add required files
            to_serve.append(
                _create_test(f"test_{i:0>3d}.html", tmp_path, data=b"A" * ((i % 2) + 1))
            )
            # add a missing files
            to_serve.append(_TestFile(f"missing_{i:0>3d}.html"))
            # add optional files
            opt_path = tmp_path / f"opt_{i:0>3d}.html"
            opt_path.write_bytes(b"A" * ((i % 2) + 1))
            to_serve.append(_TestFile(opt_path.name))
            # add redirects
            redir_target = _create_test(f"redir_{i:0>3d}.html", tmp_path, data=b"AA")
            to_serve.append(_TestFile(f"redir_{i:0>3d}"))
            smap.set_redirect(
                to_serve[-1].file, redir_target.file, required=random.getrandbits(1) > 0
            )
            # add dynamic responses
            to_serve.append(_TestFile(f"dynm_{i:0>3d}"))
            smap.set_dynamic_response(
                to_serve[-1].file, _dyn_test_cb, mime_type="text/plain"
            )
        clients = []
        for _ in range(100):  # number of clients to spawn
            clients.append(client_factory(rx_size=1))
            throttle = 0.05 if random.getrandbits(1) else 0
            clients[-1].launch("127.0.0.1", serv.port, to_serve, throttle=throttle)
        assert serv.serve_path(tmp_path, server_map=smap)[0] == Served.ALL


def test_sapphire_25(client, tmp_path):
    """test dynamic response with bad callbacks"""
    test_dr = _TestFile("dynm_test")
    smap = ServerMap()
    smap.set_dynamic_response("dynm_test", lambda _: None, mime_type="text/plain")
    test = _create_test("test_case.html", tmp_path)
    with Sapphire(timeout=10) as serv:
        client.launch("127.0.0.1", serv.port, [test_dr, test], in_order=True)
        with raises(TypeError):
            serv.serve_path(tmp_path, server_map=smap)


def test_sapphire_26(client, tmp_path):
    """test serving to a slow client"""
    t_data = "".join(random.choices("ABCD1234", k=0x19000)).encode("ascii")  # 100KB
    t_file = _create_test("test_case.html", tmp_path, data=t_data, calc_hash=True)
    # rx_size 10KB and throttle to 0.25 sec, which will be ~50KB/s
    # also taking 2.5 seconds to complete will hopefully find problems
    # with any assumptions that were made
    client.rx_size = 0x2800
    with Sapphire(timeout=60) as serv:
        client.launch("127.0.0.1", serv.port, [t_file], throttle=0.25)
        assert serv.serve_path(tmp_path)[0] == Served.ALL
    assert client.wait(timeout=10)
    assert t_file.code == 200
    assert t_file.len_srv == t_file.len_org
    assert t_file.md5_srv == t_file.md5_org


def test_sapphire_27(client, tmp_path):
    """test timeout while requesting multiple files"""
    t_data = "".join(random.choices("ABCD1234", k=1024)).encode("ascii")
    files_to_serve = [
        _create_test(f"test_{i:0>3d}.html", tmp_path, data=t_data) for i in range(50)
    ]
    client.rx_size = 512
    with Sapphire(timeout=1) as serv:  # minimum timeout is 1 second
        client.launch(
            "127.0.0.1", serv.port, files_to_serve, indicate_failure=True, throttle=0.1
        )
        status, files_served = serv.serve_path(tmp_path)
    assert status == Served.TIMEOUT
    assert len(files_served) < len(files_to_serve)


def test_sapphire_28(client_factory, tmp_path):
    """test Sapphire.serve_path() with forever=True"""
    with Sapphire(timeout=10) as serv:
        assert serv.timeout == 10
        test = _create_test("test_case.html", tmp_path)
        clients = [client_factory() for _ in range(3)]
        for client in clients:
            client.launch("127.0.0.1", serv.port, [test], skip_served=False)

        def _test_callback():
            with test.lock:
                return test.requested < 3

        assert (
            serv.serve_path(tmp_path, continue_cb=_test_callback, forever=True)[0]
            == Served.ALL
        )
    for client in clients:
        assert client.wait(timeout=10)
        client.close()
    assert test.requested == 3
    assert test.code == 200
    assert test.len_srv == test.len_org


@mark.parametrize(
    "file_name",
    [
        # space in file name
        "test case.html",
        # non-alphanumeric chars (valid characters to use on filesystem)
        "!@$%^&(_+-=[]),'~`{}",
        # extended ascii chars
        "€d’é-ñÿ",
    ],
)
def test_sapphire_29(client, tmp_path, file_name):
    """test interesting file names"""
    to_serve = [_create_test(file_name, tmp_path)]
    with Sapphire(timeout=10) as serv:
        client.launch("127.0.0.1", serv.port, to_serve)
        assert serv.serve_path(tmp_path)[0] == Served.ALL
    assert client.wait(timeout=10)
    assert all(t_file.code == 200 for t_file in to_serve)


def test_sapphire_30(client, tmp_path):
    """test interesting path string"""
    path = "".join(chr(i) for i in range(256))
    to_serve = [
        # should not trigger crash
        _TestFile(path),
        # used to keep server running
        _create_test("a.html", tmp_path),
    ]
    with Sapphire(timeout=10) as serv:
        client.launch("127.0.0.1", serv.port, to_serve, in_order=True)
        assert serv.serve_path(tmp_path, optional_files=[path])[0] == Served.ALL
    assert client.wait(timeout=10)
    assert all(t_file.code is not None for t_file in to_serve)


def test_sapphire_31(mocker):
    """test Sapphire._create_listening_socket()"""
    fake_sleep = mocker.patch("sapphire.core.sleep", autospec=True)
    fake_sock = mocker.patch("sapphire.core.socket", autospec=True)
    assert Sapphire._create_listening_socket(False, None)
    assert fake_sock.return_value.close.call_count == 0
    assert fake_sock.return_value.setsockopt.call_count == 1
    assert fake_sock.return_value.settimeout.call_count == 1
    assert fake_sock.return_value.bind.call_count == 1
    assert fake_sock.return_value.listen.call_count == 1
    assert fake_sleep.call_count == 0
    fake_sock.reset_mock()
    # failure to bind
    fake_sock.return_value.bind.side_effect = OSError
    with raises(OSError):
        Sapphire._create_listening_socket(False, None)
    assert fake_sock.return_value.close.call_count == 1
    assert fake_sleep.call_count == 0
    fake_sock.reset_mock()
    # failure and pass on retry
    exc = OSError()
    exc.errno = 10013
    fake_sock.return_value.bind.side_effect = (exc, None)
    assert Sapphire._create_listening_socket(False, None)
    assert fake_sock.return_value.close.call_count == 1
    assert fake_sock.return_value.listen.call_count == 1
    assert fake_sleep.call_count == 1


def test_sapphire_32(mocker):
    """test Sapphire.clear_backlog()"""
    mocker.patch("sapphire.core.socket", autospec=True)
    mocker.patch("sapphire.core.time", autospec=True, return_value=1)
    pending = mocker.Mock(spec_set=socket.socket)
    with Sapphire(timeout=10) as serv:
        serv._socket = mocker.Mock(spec_set=socket.socket)
        serv._socket.accept.side_effect = ((pending, None), OSError, BlockingIOError)
        serv.clear_backlog()
        assert serv._socket.accept.call_count == 3
        assert serv._socket.settimeout.call_count == 2
    assert pending.close.call_count == 1


@mark.skipif(system() != "Windows", reason="Only supported on Windows")
def test_sapphire_33(client, tmp_path):
    """test serving from path using Windows short file name"""
    wwwroot = tmp_path / "long_path_name_that_can_be_truncated_on_windows"
    wwwroot.mkdir()
    with Sapphire(timeout=10) as serv:
        assert serv.timeout == 10
        test = _create_test("test_case.html", wwwroot)
        client.launch("127.0.0.1", serv.port, [test])
        assert serv.serve_path(tmp_path / "LONG_P~1")[0] == Served.ALL
    assert client.wait(timeout=10)
    assert test.code == 200
    assert test.len_srv == test.len_org


def test_main_01(mocker, tmp_path):
    """test Sapphire.main()"""
    args = mocker.Mock(path=tmp_path, port=4536, remote=False, timeout=0)
    fake_srv = mocker.patch("sapphire.core.Sapphire.serve_path", autospec=True)
    fake_srv.return_value = (Served.ALL, None)
    Sapphire.main(args)
    fake_srv.return_value = (Served.NONE, None)
    Sapphire.main(args)
    fake_srv.side_effect = KeyboardInterrupt
    Sapphire.main(args)



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
from abc import ABCMeta, abstractmethod
from os import remove

__all__ = ("TargetMonitor",)
__author__ = "Tyson Smith"
__credits__ = ["Tyson Smith", "Jesse Schwartzentruber"]


class TargetMonitor(metaclass=ABCMeta):
    @abstractmethod
    def clone_log(self, log_id, offset=0):
        pass

    @abstractmethod
    def is_healthy(self):
        pass

    @abstractmethod
    def is_running(self):
        pass

    @property
    @abstractmethod
    def launches(self):
        pass

    def log_data(self, log_id, offset=0):
        data = None
        log_file = self.clone_log(log_id, offset=offset)
        if log_file:
            try:
                with open(log_file, "rb") as log_fp:
                    data = log_fp.read()
            finally:
                remove(log_file)
        return data

    @abstractmethod
    def log_length(self, log_id):
        pass



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
# pylint: disable=protected-access
from itertools import count
from os.path import isfile
from platform import system

from ffpuppet import BrowserTerminatedError, BrowserTimeoutError, Debugger, Reason
from pytest import mark, raises

from .puppet_target import PuppetTarget
from .target import AssetManager, Result, TargetLaunchError, TargetLaunchTimeout


def test_puppet_target_01(mocker, tmp_path):
    """test creating a PuppetTarget"""
    fake_ffp = mocker.patch("grizzly.target.puppet_target.FFPuppet", autospec=True)
    fake_ffp.return_value.reason = Reason.CLOSED
    fake_ffp.return_value.log_length.return_value = 562
    fake_file = tmp_path / "fake"
    fake_file.touch()
    with PuppetTarget(str(fake_file), 300, 25, 5000) as target:
        assert target.assets
        assert target.closed
        assert target.launch_timeout == 300
        assert target.log_limit == 25
        assert target.memory_limit == 5000
        assert target.check_result([]) == Result.NONE
        assert target.log_size() == 1124
        fake_ffp.return_value.log_length.assert_any_call("stderr")
        fake_ffp.return_value.log_length.assert_any_call("stdout")
        assert target.monitor is not None
        target.save_logs("fake_dest")
        assert fake_ffp.return_value.save_logs.call_count == 1
    assert fake_ffp.return_value.clean_up.call_count == 1
    # with extra args
    with PuppetTarget(str(fake_file), 1, 1, 1, rr=True, fake=1) as target:
        pass


def test_puppet_target_02(mocker, tmp_path):
    """test PuppetTarget.launch()"""
    fake_ffp = mocker.patch("grizzly.target.puppet_target.FFPuppet", autospec=True)
    fake_file = tmp_path / "fake"
    fake_file.touch()
    # test providing prefs.js
    with PuppetTarget(str(fake_file), 300, 25, 5000) as target:
        # launch success
        target.launch("launch_target_page")
        assert fake_ffp.return_value.launch.call_count == 1
        assert fake_ffp.return_value.close.call_count == 0
        target.close()
        # launch timeout
        fake_ffp.reset_mock()
        fake_ffp.return_value.launch.side_effect = BrowserTimeoutError("timeout")
        with raises(TargetLaunchTimeout, match="timeout"):
            target.launch("launch_target_page")
        assert fake_ffp.return_value.save_logs.call_count == 0
        # launch failure
        fake_ffp.reset_mock()
        (tmp_path / "log_stderr.txt").write_text("fake log")
        (tmp_path / "log_stdout.txt").write_text("fake log")
        mocker.patch(
            "grizzly.target.puppet_target.mkdtemp",
            autospec=True,
            return_value=str(tmp_path),
        )
        fake_ffp.return_value.launch.side_effect = BrowserTerminatedError("fail")
        with raises(TargetLaunchError, match="fail"):
            target.launch("launch_target_page")
        assert fake_ffp.return_value.save_logs.call_count == 1


@mark.parametrize(
    "healthy, reason, ignore, result, closes",
    [
        # running as expected - no failures
        (True, None, [], Result.NONE, 0),
        # browser process closed
        (False, Reason.CLOSED, [], Result.NONE, 1),
        # browser process crashed
        (False, Reason.ALERT, [], Result.FOUND, 1),
        # browser exit with no crash logs
        (False, Reason.EXITED, [], Result.NONE, 1),
        # ffpuppet check failed
        (False, Reason.WORKER, [], Result.FOUND, 1),
        # ffpuppet check ignored (memory)
        (False, Reason.WORKER, ["memory"], Result.IGNORED, 1),
        # ffpuppet check ignored (log-limit)
        (False, Reason.WORKER, ["log-limit"], Result.IGNORED, 1),
    ],
)
def test_puppet_target_03(mocker, tmp_path, healthy, reason, ignore, result, closes):
    """test PuppetTarget.check_result()"""
    fake_ffp = mocker.patch("grizzly.target.puppet_target.FFPuppet", autospec=True)
    fake_file = tmp_path / "fake"
    fake_file.touch()
    if "memory" in ignore:
        fake_ffp.return_value.available_logs.return_value = "ffp_worker_memory_usage"
    elif "log-limit" in ignore:
        fake_ffp.return_value.available_logs.return_value = "ffp_worker_log_size"
    fake_ffp.return_value.is_healthy.return_value = healthy
    fake_ffp.return_value.reason = reason
    with PuppetTarget(str(fake_file), 300, 25, 5000) as target:
        assert target.check_result(ignore) == result
    assert fake_ffp.return_value.close.call_count == closes


@mark.parametrize(
    "healthy, usage, os_name, killed",
    [
        # skip sending SIGABRT on unsupported OSs
        (True, [(1234, 90)], "Windows", 0),
        # skip idle check if target is in a bad state
        (False, [], "Linux", 0),
        # send SIGABRT to hung process
        (True, [(234, 10), (236, 75), (238, 60)], "Linux", 1),
        # ignore idle timeout (close don't abort)
        (True, [(234, 10)], "Linux", 0),
    ],
)
def test_puppet_target_04(mocker, tmp_path, healthy, usage, os_name, killed):
    """test PuppetTarget.handle_hang()"""
    mocker.patch(
        "grizzly.target.puppet_target.system", autospec=True, return_value=os_name
    )
    fake_ffp = mocker.patch("grizzly.target.puppet_target.FFPuppet", autospec=True)
    fake_kill = mocker.patch("grizzly.target.puppet_target.kill", autospec=True)
    # raise OSError for code coverage
    fake_kill.side_effect = OSError
    fake_file = tmp_path / "fake"
    fake_file.touch()
    fake_ffp.return_value.cpu_usage.return_value = usage
    fake_ffp.return_value.is_healthy.return_value = healthy
    with PuppetTarget(str(fake_file), 300, 25, 5000) as target:
        target.handle_hang()
    assert fake_ffp.return_value.is_healthy.call_count == 1
    assert fake_ffp.return_value.close.call_count == 1
    assert fake_ffp.return_value.cpu_usage.call_count == (1 if usage else 0)
    assert fake_kill.call_count == fake_ffp.return_value.wait.call_count == killed


@mark.skipif(system() != "Linux", reason="Linux only")
def test_puppet_target_05(mocker, tmp_path):
    """test PuppetTarget.dump_coverage() - full test"""
    mocker.patch("grizzly.target.puppet_target.wait_procs", autospec=True)
    fake_ffp = mocker.patch("grizzly.target.puppet_target.FFPuppet", autospec=True)
    child_proc = mocker.Mock(pid=101)
    child_proc.exe.return_value = "firefox-bin"
    fake_proc = mocker.patch("grizzly.target.puppet_target.Process", autospec=True)
    fake_proc.return_value.pid = 100
    fake_proc.return_value.children.return_value = (child_proc,)
    fake_proc.return_value.exe.return_value = "firefox-bin"
    fake_proc_iter = mocker.patch(
        "grizzly.target.puppet_target.process_iter", autospec=True
    )
    mocker.patch("grizzly.target.puppet_target.sleep", autospec=True)
    fake_time = mocker.patch("grizzly.target.puppet_target.time", autospec=True)
    fake_file = tmp_path / "fake"
    fake_file.touch()
    target = PuppetTarget(str(fake_file), 300, 25, 5000)
    fake_kill = mocker.patch("grizzly.target.puppet_target.kill", autospec=True)
    # not running
    fake_ffp.return_value.get_pid.return_value = None
    target.dump_coverage()
    assert not fake_kill.call_count
    assert fake_ffp.return_value.get_pid.call_count == 1
    assert fake_proc_iter.call_count == 0
    # gcda not found
    fake_ffp.return_value.is_healthy.return_value = True
    fake_ffp.return_value.get_pid.return_value = 100
    fake_time.side_effect = (0, 1, 10)
    target.dump_coverage()
    assert fake_kill.call_count == 2
    assert fake_proc_iter.call_count == 2
    assert fake_ffp.return_value.is_healthy.call_count == 2
    fake_ffp.reset_mock()
    fake_kill.reset_mock()
    fake_proc_iter.reset_mock()
    # browser crashes
    fake_ffp.return_value.is_healthy.side_effect = (True, False)
    fake_time.side_effect = None
    fake_time.return_value = 1.0
    target.dump_coverage()
    assert fake_kill.call_count == 2
    assert fake_proc_iter.call_count == 1
    assert fake_ffp.return_value.is_healthy.call_count == 2
    fake_ffp.reset_mock()
    fake_kill.reset_mock()
    fake_proc_iter.reset_mock()
    # timeout while waiting for files
    fake_ffp.return_value.is_healthy.return_value = True
    fake_ffp.return_value.is_healthy.side_effect = None
    fake_ffp.return_value.get_pid.return_value = 100
    fake_proc_iter.return_value = (
        mocker.Mock(info={"pid": 101, "open_files": (mocker.Mock(path="a.gcda"),)}),
    )
    fake_time.side_effect = (0, 1, 20)
    target.dump_coverage(timeout=15)
    assert fake_kill.call_count == 3
    assert fake_proc_iter.call_count == 2
    assert fake_ffp.return_value.is_healthy.call_count == 2
    assert fake_ffp.return_value.close.call_count == 1
    fake_ffp.reset_mock()
    fake_kill.reset_mock()
    fake_proc_iter.reset_mock()
    # wait for files (success)
    fake_ffp.return_value.get_pid.return_value = 100
    fake_time.side_effect = None
    fake_time.return_value = 1.0
    fake_proc_iter.side_effect = (
        # 1st call
        (
            mocker.Mock(
                info={
                    "pid": 100,
                    "open_files": (
                        mocker.Mock(path="a.bin"),
                        mocker.Mock(path="/a/s/d"),
                    ),
                }
            ),
            mocker.Mock(info={"pid": 101, "open_files": None}),
            mocker.Mock(info={"pid": 999, "open_files": None}),
        ),
        # 2nd call
        (mocker.Mock(info={"pid": 100, "open_files": (mocker.Mock(path="a.gcda"),)}),),
        # 3rd call
        (
            mocker.Mock(info={"pid": 100, "open_files": (mocker.Mock(path="a.bin"),)}),
            mocker.Mock(
                info={"pid": 999, "open_files": (mocker.Mock(path="ignore.gcda"),)}
            ),
        ),
    )
    target.dump_coverage()
    assert fake_ffp.return_value.close.call_count == 0
    assert fake_proc_iter.call_count == 3
    assert fake_kill.call_count == 2
    fake_ffp.reset_mock()
    fake_kill.reset_mock()
    fake_proc_iter.reset_mock()
    # kill calls raise OSError
    fake_kill.side_effect = OSError
    fake_ffp.return_value.is_healthy.return_value = True
    fake_ffp.return_value.get_pid.return_value = 100
    fake_proc_iter.side_effect = None
    fake_time.side_effect = count()
    target.dump_coverage()
    assert fake_kill.call_count == 2
    fake_ffp.reset_mock()
    fake_kill.reset_mock()
    fake_proc_iter.reset_mock()
    target.cleanup()


def test_puppet_target_06(mocker, tmp_path):
    """test PuppetTarget.is_idle()"""
    fake_ffp = mocker.patch("grizzly.target.puppet_target.FFPuppet", autospec=True)
    fake_ffp.return_value.cpu_usage.return_value = [(999, 30), (998, 20), (997, 10)]
    fake_file = tmp_path / "fake"
    fake_file.touch()
    with PuppetTarget(str(fake_file), 300, 25, 5000) as target:
        assert not target.is_idle(0)
        assert not target.is_idle(25)
        assert target.is_idle(50)


def test_puppet_target_07(mocker, tmp_path):
    """test PuppetTarget.monitor"""
    fake_ffp = mocker.patch("grizzly.target.puppet_target.FFPuppet", autospec=True)
    fake_file = tmp_path / "fake"
    fake_file.touch()
    with PuppetTarget(str(fake_file), 300, 25, 5000) as target:
        fake_ffp.return_value.is_running.return_value = False
        fake_ffp.return_value.is_healthy.return_value = False
        assert target.monitor is not None
        assert not target.monitor.is_healthy()
        assert not target.monitor.is_running()
        fake_ffp.return_value.is_running.return_value = True
        fake_ffp.return_value.is_healthy.return_value = True
        assert target.monitor.is_healthy()
        assert target.monitor.is_running()
        fake_ffp.return_value.launches = 123
        assert target.monitor.launches == 123
        fake_ffp.return_value.log_length.return_value = 100
        assert target.monitor.log_length("stdout") == 100
        target.monitor.clone_log("somelog")
        assert fake_ffp.return_value.clone_log.call_count == 1


def test_puppet_target_08(mocker, tmp_path):
    """test PuppetTarget.process_assets()"""
    mocker.patch("grizzly.target.puppet_target.FFPuppet", autospec=True)
    fake_file = tmp_path / "fake"
    fake_file.write_text("1\n2\n")
    # no prefs file provided
    with PuppetTarget(str(fake_file), 300, 25, 5000) as target:
        assert target.assets.get("prefs") is None
        target.process_assets()
        assert isfile(target.assets.get("prefs"))
        assert target.assets.get("prefs").endswith("prefs.js")
    # prefs file provided
    with AssetManager(base_path=str(tmp_path)) as assets:
        assets.add("prefs", str(fake_file))
        with PuppetTarget(str(fake_file), 300, 25, 5000, assets=assets) as target:
            target.process_assets()
            assert isfile(target.assets.get("prefs"))
            assert target.assets.get("prefs").endswith("fake")
    # abort tokens file provided
    with AssetManager(base_path=str(tmp_path)) as assets:
        assets.add("abort-tokens", str(fake_file))
        with PuppetTarget(str(fake_file), 300, 25, 5000, assets=assets) as target:
            # ignore E1101: (pylint 2.9.3 bug?)
            #    Method 'add_abort_token' has no 'call_count' member (no-member)
            # pylint: disable=no-member
            assert target._puppet.add_abort_token.call_count == 0
            target.process_assets()
            assert isfile(target.assets.get("abort-tokens"))
            assert target.assets.get("abort-tokens").endswith("fake")
            assert target._puppet.add_abort_token.call_count == 2


@mark.parametrize(
    "pernosco, rr, valgrind",
    [
        # No debugger selected
        (False, False, False),
        # Pernosco selected
        (True, False, False),
        # rr selected
        (False, True, False),
        # Valgrind selected
        (False, False, True),
    ],  # pylint: disable=invalid-name
)
def test_puppet_target_09(
    mocker, tmp_path, pernosco, rr, valgrind
):  # pylint: disable=invalid-name
    """test PuppetTarget debugger args"""
    fake_ffp = mocker.patch("grizzly.target.puppet_target.FFPuppet", autospec=True)
    fake_file = tmp_path / "fake"
    fake_file.touch()
    with PuppetTarget(
        str(fake_file), 30, 25, 500, pernosco=pernosco, rr=rr, valgrind=valgrind
    ) as _:
        pass
    if pernosco:
        assert fake_ffp.call_args[-1]["debugger"] == Debugger.PERNOSCO
    elif rr:
        assert fake_ffp.call_args[-1]["debugger"] == Debugger.RR
    elif valgrind:
        assert fake_ffp.call_args[-1]["debugger"] == Debugger.VALGRIND
    else:
        assert fake_ffp.call_args[-1]["debugger"] == Debugger.NONE


@mark.parametrize(
    "asset, env",
    [
        # suppressions via asset
        (True, False),
        # suppressions via env
        (False, True),
        # suppressions via both asset and env (asset should be preferred)
        (True, True),
        # missing suppressions file
        (False, False),
    ],
)
def test_puppet_target_10(tmp_path, asset, env):
    """test PuppetTarget.process_assets() - configure sanitizer suppressions"""
    fake_file = tmp_path / "fake"
    fake_file.touch()
    supp_asset = tmp_path / "supp_asset"
    supp_env = tmp_path / "supp_env"
    with AssetManager(base_path=str(tmp_path)) as assets:
        assets.add("prefs", str(fake_file))
        if asset:
            supp_asset.touch()
            assets.add("lsan-suppressions", str(supp_asset))
        with PuppetTarget(str(fake_file), 300, 25, 5000, assets=assets) as target:
            target.environ["TSAN_OPTIONS"] = "a=1"
            if env:
                supp_env.touch()
                target.environ["LSAN_OPTIONS"] = f"suppressions='{supp_env}'"
            else:
                target.environ["LSAN_OPTIONS"] = "suppressions='missing'"
            target.process_assets()
            if asset:
                assert supp_asset.name in target.environ["LSAN_OPTIONS"]
            elif env:
                assert supp_env.name in target.environ["LSAN_OPTIONS"]
                assert assets.get("lsan-suppressions")
            else:
                assert not assets.get("lsan-suppressions")


def test_puppet_target_11(tmp_path):
    """test PuppetTarget.filtered_environ()"""
    fake_file = tmp_path / "fake"
    fake_file.touch()
    with PuppetTarget(str(fake_file), 300, 25, 5000) as target:
        target.environ = {
            "TRACKED": "1",
            "ASAN_OPTIONS": "external_symbolizer_path='a':no_remove='b'",
            "LSAN_OPTIONS": "suppressions='a'",
            "EMPTY": "",
        }
        assert "EMPTY" not in target.filtered_environ()
        assert "TRACKED" in target.filtered_environ()
        assert "ASAN_OPTIONS" in target.filtered_environ()
        assert "LSAN_OPTIONS" not in target.filtered_environ()


@mark.parametrize(
    "base, extra, result",
    [
        # empty
        ({}, {}, {}),
        # empty extra
        ({"a": "1"}, {}, {"a": "1"}),
        # empty base
        ({}, {"a": "1"}, {"a": "1"}),
        # merge
        ({"a": "1"}, {"b": "2"}, {"a": "1", "b": "2"}),
        # name collision, favor base
        ({"a": "1"}, {"a": "2"}, {"a": "1"}),
        # name collision and merge
        ({"a": "1"}, {"a": "2", "b": "2"}, {"a": "1", "b": "2"}),
    ],
)
def test_puppet_target_12(tmp_path, base, extra, result):
    """test PuppetTarget.merge_environment()"""
    fake_file = tmp_path / "fake"
    fake_file.touch()
    with PuppetTarget(str(fake_file), 300, 25, 5000) as target:
        target.environ = base
        target.merge_environment(extra)
        assert target.environ == result


@mark.parametrize(
    "base, extra, result",
    [
        # collision with existing option
        ({"ASAN_OPTIONS": "a=1"}, {"ASAN_OPTIONS": "a=2"}, ["a=1"]),
        # add option from extra
        ({"ASAN_OPTIONS": "a=1"}, {"ASAN_OPTIONS": "b=2:c3"}, ["a=1", "b=2", "c=3"]),
        # add option from extra
        ({"ASAN_OPTIONS": "a=1:c=3"}, {"ASAN_OPTIONS": "b=2"}, ["a=1", "b=2", "c=3"]),
    ],
)
def test_puppet_target_13(tmp_path, base, extra, result):
    """test PuppetTarget.merge_environment() - merge sanitizer options"""
    fake_file = tmp_path / "fake"
    fake_file.touch()
    with PuppetTarget(str(fake_file), 300, 25, 5000) as target:
        target.environ = base
        target.merge_environment(extra)
        for opt in target.environ["ASAN_OPTIONS"].split(":"):
            assert opt in result


def test_puppet_target_14(mocker, tmp_path):
    """test PuppetTarget.dump_coverage() - skip on unsupported platform"""
    mocker.patch("grizzly.target.puppet_target.system", return_value="foo")
    fake_file = tmp_path / "fake"
    fake_file.touch()
    with PuppetTarget(str(fake_file), 300, 25, 5000) as target:
        target.dump_coverage()



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
from abc import ABCMeta, abstractmethod
from enum import IntEnum, unique
from json import dumps, loads
from logging import getLogger
from os import getenv
from pathlib import Path
from shutil import copyfile, move, rmtree
from tempfile import TemporaryDirectory
from zipfile import ZIP_DEFLATED, ZipFile

# import FuzzManager utilities
from Collector.Collector import Collector
from fasteners.process_lock import InterProcessLock
from FTB.ProgramConfiguration import ProgramConfiguration
from psutil import disk_usage

from .report import Report
from .utils import grz_tmp

__all__ = ("FilesystemReporter", "FuzzManagerReporter", "Quality")
__author__ = "Tyson Smith"
__credits__ = ["Tyson Smith"]

LOG = getLogger(__name__)


@unique
class Quality(IntEnum):
    """testcase quality values"""

    # final reduced testcase
    REDUCED = 0
    # original used for reduction (a reduced version exists)
    ORIGINAL = 1
    # used to manually mark a report as unreducible
    IGNORED = 3
    # the testcase is currently being reduced
    REDUCING = 4
    # haven't attempted reduction yet (1st attempt, generic reducer)
    UNREDUCED = 5
    # platform specific reduction requested (2nd attempt)
    REQUEST_SPECIFIC = 6
    # testcase not detected ("testcase" not a testcase?)
    NO_TESTCASE = 7
    # the testcase was reproducible, but broke during reduction
    REDUCER_BROKE = 8
    # reducer error
    REDUCER_ERROR = 9
    # could not reproduce the testcase
    NOT_REPRODUCIBLE = 10


class Reporter(metaclass=ABCMeta):
    __slots__ = ("display_logs",)

    def __init__(self):
        self.display_logs = getenv("GRZ_DISPLAY_REPORT") == "1"

    @abstractmethod
    def _post_submit(self):
        pass

    @abstractmethod
    def _pre_submit(self, report):
        pass

    @abstractmethod
    def _submit_report(self, report, test_cases):
        pass

    def submit(self, test_cases, report):
        """Submit report containing results.

        Args:
            test_cases (iterable): A collection of testcases, ordered newest to oldest,
                                   the newest being the mostly likely to trigger
                                   the result (crash, assert... etc).
            report (Report): Report to submit.

        Returns:
            *: implementation specific result indicating where the report was created
        """
        assert isinstance(report, Report)
        assert report.path is not None
        self._pre_submit(report)
        # output report contents to console
        if self.display_logs:
            if not report.is_hang:
                LOG.info(
                    "=== BEGIN REPORT ===\n%s",
                    report.preferred.read_text("utf-8", errors="ignore"),
                )
            else:
                LOG.info("=== BEGIN REPORT ===\nBrowser hang detected")
            LOG.info("=== END REPORT ===")
        result = self._submit_report(report, test_cases)
        if report is not None:
            report.cleanup()
        self._post_submit()
        return result


class FilesystemReporter(Reporter):
    DISK_SPACE_ABORT = 512 * 1024 * 1024  # 512 MB

    __slots__ = ("major_bucket", "min_space", "report_path")

    def __init__(self, report_path, major_bucket=True):
        super().__init__()
        self.major_bucket = major_bucket
        self.min_space = FilesystemReporter.DISK_SPACE_ABORT
        self.report_path = Path(report_path)

    def _pre_submit(self, report):
        pass

    def _post_submit(self):
        pass

    def _submit_report(self, report, test_cases):
        # create major bucket directory in working directory if needed
        if self.major_bucket:
            dest = self.report_path / report.major[:16]
        else:
            dest = self.report_path
        dest.mkdir(parents=True, exist_ok=True)
        # dump test cases and the contained files to working directory
        for test_number, test_case in enumerate(test_cases):
            dump_path = dest / f"{report.prefix}-{test_number}"
            dump_path.mkdir(exist_ok=True)
            test_case.dump(dump_path, include_details=True)
        # move logs into bucket directory
        log_path = dest / f"{report.prefix}_logs"
        if log_path.is_dir():
            LOG.warning("Report log path exists '%s'", log_path)
        move(report.path, log_path)
        # avoid filling the disk
        free_space = disk_usage(str(log_path)).free
        if free_space < self.min_space:
            raise RuntimeError(f"Low disk space: {free_space / 1_048_576:0.1f}MBs")
        return dest


class FailedLaunchReporter(FilesystemReporter):
    """Save launch failure reports to disk."""

    def __init__(self, display=False):
        super().__init__(grz_tmp("launch_failures"), major_bucket=False)
        self.display_logs = display
        if not display:
            LOG.info("Display logs with --display-launch-failures")

    def _post_submit(self):
        super()._post_submit()
        LOG.info("Logs can be found here '%s'", self.report_path)


class FuzzManagerReporter(Reporter):
    FM_CONFIG = Path.home() / ".fuzzmanagerconf"

    __slots__ = ("_extra_metadata", "force_report", "quality", "tool")

    def __init__(self, tool=None):
        super().__init__()
        self._extra_metadata = {}
        self.force_report = False
        self.quality = Quality.UNREDUCED
        self.tool = tool  # optional tool name

    def _post_submit(self):
        self._extra_metadata.clear()

    @staticmethod
    def sanity_check(bin_file):
        """Perform FuzzManager sanity check.

        Args:
            bin_file (str): Binary file being tested.

        Returns:
            None
        """
        if not FuzzManagerReporter.FM_CONFIG.is_file():
            raise OSError(f"Missing: {FuzzManagerReporter.FM_CONFIG}")
        if not Path(f"{bin_file}.fuzzmanagerconf").is_file():
            raise OSError(f"Missing: {bin_file}.fuzzmanagerconf")
        ProgramConfiguration.fromBinary(bin_file)

    def add_extra_metadata(self, key, value):
        """Add extra metadata to be reported with any CrashEntrys reported.

        Arguments:
            key (str): key for this data in the metadata dict
            value (object): JSON serializable object to be included in the FM crash
                            metadata. The object will be deep-copied.

        Returns:
            None
        """
        assert isinstance(key, str)
        assert key not in self._extra_metadata
        # deep copy and ensure that value is JSON serializable
        self._extra_metadata[key] = loads(dumps(value))

    def _pre_submit(self, report):
        self._process_rr_trace(report)

    def _process_rr_trace(self, report):
        # don't report large files to FuzzManager
        trace_path = report.path / "rr-traces"
        if trace_path.is_dir():
            LOG.info("Ignored rr trace")
            self.add_extra_metadata("rr-trace", "ignored")
            # remove traces so they are not uploaded to FM (because they are huge)
            rmtree(trace_path)

    @staticmethod
    def _ignored(report):
        # This is here to prevent reporting stack-less crashes
        # that were caused by system OOM
        log_data = report.preferred.read_text("utf-8", errors="ignore")
        # ignore sanitizer OOMs missing stack
        if report.stack is None:
            mem_errs = (
                "ERROR: Failed to mmap",
                # NOTE: max_allocation_size_mb can trigger a similar message
                ": AddressSanitizer failed to allocate",
                "Sanitizer: internal allocator is out of memory trying to allocate",
            )
            # scan log data for memory error strings
            if any(msg in log_data for msg in mem_errs):
                return True
        # ignore Valgrind crashes
        return log_data.startswith("VEX temporary storage exhausted.")

    def _submit_report(self, report, test_cases):
        collector = Collector()

        if not self.force_report:
            if collector.sigCacheDir and Path(collector.sigCacheDir).is_dir():
                # search for a cached signature match
                with InterProcessLock(str(grz_tmp() / "fm_sigcache.lock")):
                    _, cache_metadata = collector.search(report.crash_info)
                # check if signature has been marked as frequent in FM
                if cache_metadata and cache_metadata["frequent"]:
                    LOG.info(
                        "Frequent crash matched existing signature: %s",
                        cache_metadata["shortDescription"],
                    )
                    return None
            else:
                LOG.debug("sigCacheDir does not exist (%r)", collector.sigCacheDir)

        if self._ignored(report):
            LOG.info("Report is in ignore list")
            return None

        if report.is_hang:
            self.add_extra_metadata("is_hang", True)

        # dump test cases and the contained files to working directory
        test_case_meta = []
        for test_number, test_case in enumerate(test_cases):
            test_case_meta.append([test_case.adapter_name, test_case.input_fname])
            dump_path = report.path / f"{report.prefix}-{test_number}"
            dump_path.mkdir(exist_ok=True)
            test_case.dump(dump_path, include_details=True)
        report.crash_info.configuration.addMetadata(
            {"grizzly_input": repr(test_case_meta)}
        )
        if test_cases:
            environ_string = " ".join(
                "=".join(kv) for kv in test_cases[0].env_vars.items()
            )
            report.crash_info.configuration.addMetadata(
                {"recorded_envvars": environ_string}
            )
        else:
            self.quality = Quality.NO_TESTCASE
        report.crash_info.configuration.addMetadata(self._extra_metadata)

        # TODO: this should likely move to ffpuppet
        # grab screen log (used in automation)
        if getenv("WINDOW") is not None:
            screen_log = Path.cwd() / f"screenlog.{getenv('WINDOW')}"
            if screen_log.is_file():
                target_log = report.path / "screenlog.txt"
                copyfile(screen_log, target_log)
                Report.tail(target_log, 10240)  # limit to last 10K

        with TemporaryDirectory(prefix="fm-zip", dir=grz_tmp()) as tmp_dir:
            # add results to a zip file
            zip_name = Path(tmp_dir) / f"{report.prefix}.zip"
            with ZipFile(zip_name, mode="w", compression=ZIP_DEFLATED) as zip_fp:
                # add test files
                for entry in report.path.rglob("*"):
                    if entry.is_file():
                        zip_fp.write(entry, arcname=entry.relative_to(report.path))
            # override tool name if specified
            if self.tool is not None:
                collector.tool = self.tool

            # submit results to the FuzzManager server
            new_entry = collector.submit(
                report.crash_info, testCase=zip_name, testCaseQuality=self.quality.value
            )
        LOG.info("Logged %d (%s)", new_entry["id"], self.quality.name)

        return new_entry["id"]



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
"""Grizzly reducer CLI.
"""
from . import ReduceManager
from .args import ReduceArgs

raise SystemExit(ReduceManager.main(ReduceArgs().parse_args()))



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
from .assets import AssetManager
from .target import Target


class SimpleTarget(Target):
    TRACKED_ENVVARS = ("TEST_INC",)

    def _cleanup(self):
        pass

    def check_result(self, ignored):
        pass

    def close(self, force_close=False):
        pass

    @property
    def closed(self):
        pass

    def create_report(self, is_hang=False):
        pass

    def handle_hang(self, ignore_idle=True):
        pass

    def launch(self, location):
        pass

    @property
    def monitor(self):
        return self._monitor

    def merge_environment(self, extra):
        pass

    def process_assets(self):
        pass

    def save_logs(self, *_args, **_kwargs):
        pass


def test_target_01(tmp_path):
    """test creating a simple Target"""
    fake_file = tmp_path / "fake"
    fake_file.touch()
    with SimpleTarget(str(fake_file), 10, 2, 3) as target:
        assert target.binary == str(fake_file)
        assert target.assets
        org_path = target.assets.path
        target.assets = AssetManager(base_path=str(tmp_path))
        assert target.assets.path != org_path
        assert not target.environ
        assert not target.filtered_environ()
        assert not target.is_idle(0)
        assert target.launch_timeout == 10
        assert target.log_size() == 0
        assert target.log_limit == 2
        assert target.memory_limit == 3
        assert target.monitor is None
        # test stubs
        target.dump_coverage()
        target.reverse(1, 2)


def test_target_02(mocker, tmp_path):
    """test loading TRACKED_ENVVARS"""
    mocker.patch.dict("grizzly.target.target.environ", {"SKIP": "x", "TEST_INC": "1"})
    fake_file = tmp_path / "fake"
    fake_file.touch()
    with SimpleTarget(str(fake_file), 321, 2, 3) as target:
        assert target.environ
        assert "SKIP" not in target.environ
        assert target.environ["TEST_INC"] == "1"
        filtered = target.filtered_environ()
        assert filtered
        assert "SKIP" not in filtered
        assert filtered["TEST_INC"] == "1"


def test_target_03():
    """test Target.scan_environment()"""
    assert not Target.scan_environment({"a": "1"}, [])
    assert not Target.scan_environment({}, ["a"])
    assert Target.scan_environment({"a": "1", "b": "2"}, ["a"]) == {"a": "1"}



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
from logging import getLogger
from os import makedirs, unlink
from os.path import abspath, basename, exists, isdir, isfile
from os.path import join as pathjoin
from shutil import copyfile, copytree, move, rmtree
from tempfile import mkdtemp

__all__ = ("AssetError", "AssetManager")
__author__ = "Tyson Smith"
__credits__ = ["Tyson Smith"]

LOG = getLogger(__name__)


class AssetError(Exception):
    """Raised by AssetManager"""


class AssetManager:
    __slots__ = ("assets", "path")

    def __init__(self, base_path=None):
        self.assets = {}
        self.path = mkdtemp(prefix="assets_", dir=base_path)

    def __enter__(self):
        return self

    def __exit__(self, *exc):
        self.cleanup()

    def add(self, asset, path, copy=True):
        """Add asset to the AssetManager.

        Args:
            asset (str): Name of asset.
            path (str): Path of content to use as asset.
            copy (bool): Copy or move the content.

        Returns:
            str: Path to the asset on the filesystem.
        """
        assert isinstance(asset, str)
        assert isinstance(path, str)
        assert self.path, "cleanup() was called"
        if not exists(path):
            raise OSError(f"{path!r} does not exist")
        path = abspath(path)
        # only copy files from outside working path
        if path.startswith(self.path):
            raise AssetError(f"Cannot add existing asset content {path!r}")
        # remove existing asset with the same name
        if asset in self.assets:
            LOG.debug("asset %r exists, removing existing", asset)
            self.remove(asset)
        dst_path = pathjoin(self.path, basename(path))
        # avoid overwriting data that is part of an existing asset
        if exists(dst_path):
            raise AssetError(f"{basename(path)!r} is an existing asset")
        if copy:
            if isfile(path):
                copyfile(path, dst_path)
            else:
                copytree(path, dst_path)
        else:
            move(path, self.path)
        self.assets[asset] = dst_path
        LOG.debug(
            "added asset %r %s to %r", asset, "copied" if copy else "moved", dst_path
        )
        return dst_path

    def add_batch(self, assets):
        """Add collection of assets to the AssetManager.

        Args:
            assets (list(list(str, str))): List of list that contain asset, path pairs.

        Returns:
            None
        """
        for asset, path in assets:
            self.add(asset, path)

    def cleanup(self):
        """Remove asset files from filesystem.

        Args:
            None

        Returns:
            None
        """
        if self.path:
            rmtree(self.path, ignore_errors=True)
            self.assets.clear()
            self.path = None

    def dump(self, dst_path, subdir="_assets_"):
        """Copy assets content to a given path.

        Args:
            dst_path (str): Path to copy assets content to.
            subdir (str): Create and use as destination if given.

        Returns:
            dict: Collection asset paths keyed by asset name.
        """
        dumped = {}
        if self.assets:
            if subdir:
                dst_path = pathjoin(dst_path, subdir)
            makedirs(dst_path, exist_ok=not subdir)
            for asset, src in self.assets.items():
                dst_name = basename(src)
                dumped[asset] = dst_name
                if isfile(src):
                    copyfile(src, pathjoin(dst_path, dst_name))
                elif isdir(src):
                    copytree(src, pathjoin(dst_path, dst_name))
                else:
                    dumped.pop(asset)
                    LOG.warning("Failed to dump asset %r from %r", asset, src)
        return dumped

    def get(self, asset):
        """Get path to content on filesystem for given asset.

        Args:
            asset (str): Asset to lookup.

        Returns:
            str: Path to asset content or None if asset does not exist.
        """
        return self.assets.get(asset, None)

    def is_empty(self):
        """Check if AssetManager contains entries.

        Args:
            None

        Returns:
            bool: True if AssetManager contains entries else False.
        """
        return not self.assets

    @classmethod
    def load(cls, assets, src_path, base_path=None):
        """Load assets from filesystem.

        Args:
            asset (dict): Asset paths on filesystem relative to src_path, keyed on
                          asset name.
            src_path (str): Path to scan for assets.
            base_path (str): Base path to use to create local storage.

        Returns:
            AssetManager: Populated with contents provided by assets argument.
        """
        obj = cls(base_path=base_path)
        for asset, src_name in assets.items():
            obj.add(asset, pathjoin(src_path, src_name))
        return obj

    def remove(self, asset):
        """Remove asset from AssetManager if asset exists.

        Args:
            asset (str): Asset to remove.

        Returns:
            None
        """
        path = self.assets.pop(asset, None)
        if path:
            if isfile(path):
                unlink(path)
            else:
                rmtree(path, ignore_errors=True)



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
from os.path import isfile

from .target_monitor import TargetMonitor


def test_target_monitor_01(tmp_path):
    """test a basic TargetMonitor"""

    class _BasicMonitor(TargetMonitor):
        def clone_log(self, log_id, offset=0):
            log_file = tmp_path / "test_log.txt"
            log_file.write_bytes(b"test")
            return str(log_file)

        def is_healthy(self):
            return True

        def is_running(self):
            return True

        @property
        def launches(self):
            return 1

        def log_length(self, log_id):
            return 100

    mon = _BasicMonitor()
    test_log = mon.clone_log("test_log", offset=0)
    assert isfile(test_log)
    assert mon.is_healthy()
    assert mon.is_running()
    assert mon.launches == 1
    assert mon.log_data("test_log") == b"test"
    assert mon.log_length("test_log") == 100



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
"""
Loki fuzzing library
"""
from logging import ERROR, INFO, basicConfig, getLogger
from os import SEEK_END, makedirs
from os.path import abspath, getsize
from os.path import join as pathjoin
from os.path import splitext
from random import choice, getrandbits, randint, sample
from shutil import copy
from struct import pack, unpack
from tempfile import SpooledTemporaryFile, mkdtemp
from time import strftime, time

__author__ = "Tyson Smith"

LOG = getLogger(__name__)


class Loki:
    BYTE_ORDERS = ("<", ">", "@", "!", "=")

    __slots__ = ("aggr", "byte_order")

    def __init__(self, aggression=0.0, byte_order=None):
        self.aggr = min(max(aggression, 0.0), 1.0)
        self.byte_order = byte_order

    @staticmethod
    def _fuzz_data(in_data, byte_order):
        data_size = len(in_data)
        if data_size == 1:
            pack_unit = "B"
            mask = 0xFF
        elif data_size == 2:
            pack_unit = f"{byte_order}H"
            mask = 0xFFFF
        elif data_size == 4:
            pack_unit = f"{byte_order}I"
            mask = 0xFFFFFFFF
        else:
            raise AssertionError(f"Unsupported data size: {data_size}")

        fuzz_op = randint(0, 4)
        if fuzz_op == 0:  # boundary
            out_data = (2 ** randint(2, data_size * 8)) + randint(-2, 2)
        elif fuzz_op == 1:  # arithmetic
            out_data = unpack(pack_unit, in_data)[0] + randint(-10, 10)
        elif fuzz_op == 2:  # interesting byte, short or int
            out_data = choice((0, 1, int(mask / 2), int(mask / 2) + 1, mask))
        elif fuzz_op == 3:  # random byte, short or int
            out_data = getrandbits(32)
        elif fuzz_op == 4:
            if data_size == 1:  # toggle
                out_data = unpack(pack_unit, in_data)[0] ^ (2 ** getrandbits(3))
            elif data_size == 2:  # multiple of 8
                out_data = randint(1, 0x1FFF) * 8
            elif data_size == 4:  # multiple of 8
                out_data = randint(1, 0x1FFFFFFF) * 8

        return pack(pack_unit, out_data & mask)

    def _fuzz(self, tgt_fp):
        tgt_fp.seek(0, SEEK_END)
        length = tgt_fp.tell()
        if length < 1:
            LOG.debug("cannot fuzz empty file")
            return
        # minimum number of mutations should be 1
        max_mutations = max(int(round(length * self.aggr)), 1)
        mutations = randint(1, max_mutations)
        LOG.debug(
            "%d of a possible %d mutations will be performed", mutations, max_mutations
        )
        if self.byte_order is not None:
            assert self.byte_order in ("<", ">", "@", "!", "=")
            byte_order = self.byte_order
        else:
            LOG.debug("using random byte order")
        for count, idx in enumerate(sample(range(length), k=mutations)):
            # every few iterations randomly attempt multi-byte mutations
            if count % 10 == 0 and getrandbits(1):
                max_size = length - idx
                # target multiple bytes if possible
                if max_size > 3 and getrandbits(1):
                    size = 4
                elif max_size > 1:
                    size = 2
                else:
                    size = 1
            else:
                # target single byte
                size = 1
            if self.byte_order is None:
                byte_order = "<" if getrandbits(1) else ">"
            # perform mutation
            tgt_fp.seek(idx)
            out_data = self._fuzz_data(tgt_fp.read(size), byte_order)
            tgt_fp.seek(idx)
            tgt_fp.write(out_data)

    def fuzz_data(self, data):
        assert data
        assert isinstance(data, bytes)
        # open a temp file in memory for fuzzing
        with SpooledTemporaryFile(max_size=0x800000, mode="r+b") as tmp_fp:
            tmp_fp.write(data)
            self._fuzz(tmp_fp)
            tmp_fp.seek(0)
            return tmp_fp.read()

    def fuzz_file(self, in_file, count, out_dir, ext=None):
        try:
            if getsize(in_file) < 1:
                LOG.error("Input must be at least 1 byte long")
                return False
        except OSError:
            LOG.error("%r does not exists!", in_file)
            return False
        if ext is None:
            ext = splitext(in_file)[1]
        for i in range(count):
            out_file = pathjoin(out_dir, f"{i:0>6d}_fuzzed{ext}")
            copy(in_file, out_file)
            with open(out_file, "r+b") as out_fp:
                self._fuzz(out_fp)
        return True

    @classmethod
    def main(cls, args):
        basicConfig(format="", level=INFO if not args.quiet else ERROR)
        LOG.info("Starting Loki @ %s", strftime("%Y-%m-%d %H:%M:%S"))
        LOG.info("Target template is %r", abspath(args.input))
        out_dir = args.output
        if out_dir is None:
            out_dir = mkdtemp(prefix=strftime("loki_%m-%d_%H-%M_"), dir=".")
        makedirs(out_dir, exist_ok=True)
        LOG.info("Output directory is %r", abspath(out_dir))
        count = max(args.count, 1)
        LOG.info("Generating %d fuzzed test cases...", count)
        loki = Loki(aggression=args.aggression, byte_order=args.byte_order)
        try:
            start_time = time()
            success = loki.fuzz_file(args.input, count, out_dir)
            finish_time = time() - start_time
            LOG.info("Done. Total run time %gs", finish_time)
            if success:
                LOG.info("About %gs per file", finish_time / count)
        except KeyboardInterrupt:  # pragma: no cover
            LOG.info("Ctrl+C detected.")
            success = False
        return 0 if success else 1



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
from argparse import ArgumentParser
from logging import DEBUG, INFO, basicConfig
from pathlib import Path

from .core import Sapphire


def configure_logging(log_level):
    if log_level == DEBUG:
        date_fmt = None
        log_fmt = "%(asctime)s %(levelname).1s %(name)s | %(message)s"
    else:
        date_fmt = "%Y-%m-%d %H:%M:%S"
        log_fmt = "[%(asctime)s] %(message)s"
    basicConfig(format=log_fmt, datefmt=date_fmt, level=log_level)


def parse_args(argv=None):
    # log levels for console logging
    level_map = {"DEBUG": DEBUG, "INFO": INFO}
    parser = ArgumentParser()
    parser.add_argument("path", type=Path, help="Specify a directory to act as wwwroot")
    parser.add_argument(
        "--log-level",
        choices=sorted(level_map),
        default="INFO",
        help="Configure console logging (default: %(default)s)",
    )
    parser.add_argument(
        "--port", type=int, help="Specify a port to bind to (default: random)"
    )
    parser.add_argument(
        "--remote",
        action="store_true",
        help="Allow connections from addresses other than 127.0.0.1",
    )
    parser.add_argument(
        "--timeout",
        default=0,
        type=int,
        help="Duration in seconds to serve before exiting (default: 0 - no timeout)",
    )
    args = parser.parse_args(argv)
    # sanity check
    if not args.path.is_dir():
        parser.error(f"Path does not exist '{args.path}'")
    if args.timeout < 0:
        parser.error("--timeout must be >= 0")
    args.log_level = level_map[args.log_level]
    return args


ARGS = parse_args()
configure_logging(ARGS.log_level)
Sapphire.main(ARGS)
